import functools

import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
from dash.dependencies import Output, Input
from sqlalchemy.orm import Session
from covigator.dashboard.figures.samples import SampleFigures
from covigator.database.model import DataSource, VariantType
from covigator.database.queries import Queries

ID_VARIANTS_PER_SAMPLE_GRAPH = 'variants-per-sample-graph'
ID_INDEL_LENGTH_GRAPH = 'indel-lengths-graph'
ID_ANNOTATIONS_GRAPH = 'id-annotations-graph'
ID_SUBSTITUTIONS_GRAPH = 'substitutions-graph'
ID_SLIDER_MIN_SAMPLES = 'slider-min-samples-per-country'
ID_DROPDOWN_DATA_SOURCE = "dropdown-data-source"
ID_DROPDOWN_COUNTRY = 'dropdown-country'
ID_DROPDOWN_GENE = 'dropdown-gene-overall-mutations'
ID_DROPDOWN_GENE2 = 'dropdown-gene-overall-mutations2'
ID_DROPDOWN_VARIANT_TYPE = 'dropdown-variant-type'
ID_ACCUMULATED_SAMPLES_GRAPH = 'accumulated-samples-per-country'


@functools.lru_cache()
def get_tab_samples(queries: Queries):
    return dbc.Card(
        dbc.CardBody(
            children=[
                get_samples_tab_left_bar(queries),
                get_samples_tab_graphs()
        ])
    )


def get_samples_tab_graphs():
    return html.Div(
        className="ten columns",
        style={'overflow': 'scroll', "height": "900px"},
        children=[
            html.Br(),
            html.Div(id=ID_ACCUMULATED_SAMPLES_GRAPH),
            html.Br(),
            html.Div(children=[
                html.Div(id=ID_VARIANTS_PER_SAMPLE_GRAPH, className="five columns"),
                html.Div(id=ID_SUBSTITUTIONS_GRAPH, className="five columns"),
            ]),
            html.Div(children=[
                html.Div(id=ID_INDEL_LENGTH_GRAPH, className="five columns"),
                html.Div(id=ID_ANNOTATIONS_GRAPH, className="five columns"),
            ]),
        ])


def get_samples_tab_left_bar(queries: Queries):
    return html.Div(
        className="two columns",
        children=[
            html.Br(),
            dcc.Markdown("""Select data source"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_DATA_SOURCE,
                options=[{'label': DataSource.ENA.name, 'value': DataSource.ENA.name},
                         {'label': DataSource.GISAID.name, 'value': DataSource.GISAID.name}],
                value=None,
                multi=False
            ),
            html.Br(),
            dcc.Markdown("""Select one or more genes"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_GENE,
                options=[{'label': g.name, 'value': g.name} for g in queries.get_genes()],
                value=None,
                multi=True
            ),
            html.Br(),
            dcc.Markdown("""**Accumulated samples by country**"""),
            dcc.Markdown("""Select one or more countries"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_COUNTRY,
                options=[{'label': c, 'value': c} for c in queries.get_countries()],
                value=None,
                multi=True
            ),
            html.Br(),
            dcc.Markdown("""Minimum number of samples per country"""),
            dcc.Slider(
                id=ID_SLIDER_MIN_SAMPLES,
                min=0,
                max=10000,
                step=100,
                value=100,
                dots=False,
                tooltip=dict(always_visible=True, placement="right")
            ),
            html.Br(),
            dcc.Markdown("""**Mutations per sample and top mutations**"""),
            dcc.Markdown("""Select a variant type"""),
            dcc.Dropdown(
                id=ID_DROPDOWN_VARIANT_TYPE,
                options=[
                    {'label': VariantType.SNV.name, 'value': VariantType.SNV.name},
                    {'label': VariantType.INSERTION.name, 'value': VariantType.INSERTION.name},
                    {'label': VariantType.DELETION.name, 'value': VariantType.DELETION.name}
                ],
                value=None,
                multi=True
            ),
        ])


def set_callbacks_samples_tab(app, session: Session):

    queries = Queries(session=session)
    figures = SampleFigures(queries)

    @app.callback(
        Output(ID_ACCUMULATED_SAMPLES_GRAPH, 'children'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'),
        Input(ID_DROPDOWN_COUNTRY, 'value'),
        Input(ID_SLIDER_MIN_SAMPLES, 'value'),
        suppress_callback_exceptions=True
    )
    def update_accumulated_samples_by_country(data_source, countries, min_samples):
        return html.Div(children=figures.get_accumulated_samples_by_country_plot(
            data_source=data_source, countries=countries, min_samples=min_samples))

    @app.callback(
        Output(ID_VARIANTS_PER_SAMPLE_GRAPH, 'children'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'),
        Input(ID_DROPDOWN_GENE, 'value'),
        Input(ID_DROPDOWN_VARIANT_TYPE, 'value'),
        suppress_callback_exceptions=True
    )
    def update_variants_per_sample(data_source, genes, variant_types):
        return html.Div(children=figures.get_variants_per_sample_plot(
            data_source=data_source, genes=genes, variant_types=variant_types))

    @app.callback(
        Output(ID_SUBSTITUTIONS_GRAPH, 'children'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'),
        Input(ID_DROPDOWN_GENE, 'value'),
        Input(ID_DROPDOWN_VARIANT_TYPE, 'value'),
        suppress_callback_exceptions=True
    )
    def update_substitutions(data_source, genes, variant_types):
        return html.Div(children=figures.get_substitutions_plot(
            data_source=data_source, genes=genes, variant_types=variant_types))

    @app.callback(
        Output(ID_INDEL_LENGTH_GRAPH, 'children'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'),
        Input(ID_DROPDOWN_GENE, 'value'),
        suppress_callback_exceptions=True
    )
    def update_indel_lengths(data_source, genes):
        return html.Div(children=figures.get_indels_lengths_plot(data_source=data_source, genes=genes))

    @app.callback(
        Output(ID_ANNOTATIONS_GRAPH, 'children'),
        Input(ID_DROPDOWN_DATA_SOURCE, 'value'),
        Input(ID_DROPDOWN_GENE, 'value'),
        suppress_callback_exceptions=True
    )
    def update_annotation(data_source, genes):
        return html.Div(children=figures.get_annotations_plot(data_source=data_source, genes=genes))


from datetime import datetime

from covigator.configuration import Configuration
from covigator.database.model import JobStatus, JobGisaid, Sample, DataSource
from covigator.database.database import Database
from logzero import logger
from dask.distributed import Client

from covigator.database.queries import Queries
from covigator.processor.abstract_processor import AbstractProcessor
from covigator.pipeline.gisaid_pipeline import GisaidPipeline
from covigator.pipeline.vcf_loader import VcfLoader

MAX_INSERTIONS = 10
MAX_DELETIONS = 10
MAX_SNVS = 76


class GisaidProcessor(AbstractProcessor):

    def __init__(self, database: Database, dask_client: Client, config: Configuration):
        logger.info("Initialising GISAID processor")
        super().__init__(database, dask_client, DataSource.GISAID, config=config)

    def _process_run(self, run_accession: str):
        # NOTE: here we set the priority of each step to ensure a depth first processing
        future = self.dask_client.submit(GisaidProcessor.job, self.config, run_accession, priority=1)
        return future

    @staticmethod
    def job(config: Configuration, run_accession):
        return GisaidProcessor.run_job(
            config, run_accession, start_status=JobStatus.QUEUED, end_status=JobStatus.FINISHED,
            error_status=JobStatus.FAILED_PROCESSING, data_source=DataSource.GISAID,
            function=GisaidProcessor.run_all)

    @staticmethod
    def run_all(job: JobGisaid, queries: Queries, config: Configuration):
        GisaidProcessor.run_pipeline(job=job, queries=queries, config=config)
        GisaidProcessor.load(job=job, queries=queries, config=config)

    @staticmethod
    def run_pipeline(job: JobGisaid, queries: Queries, config: Configuration):
        sample = queries.find_sample_by_accession(job.run_accession, source=DataSource.GISAID)
        vcf = GisaidPipeline(config=config).run(sample=sample)
        job.analysed_at = datetime.now()
        job.vcf_path = vcf
        return job.run_accession

    @staticmethod
    def load(job: JobGisaid, queries: Queries, config: Configuration):
        VcfLoader().load(
            vcf_file=job.vcf_path, sample=Sample(id=job.run_accession, source=DataSource.GISAID),
            session=queries.session, max_snvs=MAX_SNVS, max_deletions=MAX_DELETIONS, max_insertions=MAX_INSERTIONS)
        job.loaded_at = datetime.now()
        return job.run_accession



import functools
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
from covigator.dashboard.figures import VARIANT_TYPE_COLOR_MAP
from covigator.dashboard.figures.figures import PLOTLY_CONFIG, MARGIN, TEMPLATE
from covigator.dashboard.tabs import get_mini_container, print_number, print_date
from covigator.database.model import DataSource, SAMPLE_ENA_TABLE_NAME, JOB_ENA_TABLE_NAME, \
    VARIANT_OBSERVATION_TABLE_NAME
from covigator.database.queries import Queries
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go


LIBRARY_STRATEGY_COLOR_MAP = {
    "WGA": "#377eb8",
    "WGS": "#4daf4a",
    "RNA-Seq": "#984ea3",
    "Targeted-Capture": "#ff7f00",
}

LIBRARY_LAYOUT_COLOR_MAP = {
    "PAIRED": "#969696",
    "SINGLE": "#cccccc"
}


@functools.lru_cache()
def get_tab_dataset_ena(queries: Queries):

    count_samples = queries.count_samples(source=DataSource.ENA.name)
    count_variants = queries.count_variant_observations(source=DataSource.ENA.name)
    date_of_first_sample = queries.get_date_of_first_sample(source=DataSource.ENA)
    date_of_most_recent_sample = queries.get_date_of_most_recent_sample(source=DataSource.ENA)

    return dbc.Card(
        dbc.CardBody(
            children=[
                dcc.Markdown("""
                The ENA dataset and its metadata was downloaded using the ENA Portal API 
                (https://www.ebi.ac.uk/ena/portal/api/). FASTQ files containing the raw reads were downloaded from the
                 provided URLs for each sample. FASTQ files were MD5 checked after download. 
                 All non-human host samples were excluded.
                 """, style={"font-size": 16}),
                html.Br(),
                html.Div(
                    html.Span(
                        children=[
                            get_mini_container(
                                title="Samples",
                                value=print_number(count_samples)
                            ),
                            get_mini_container(
                                title="Variant calls",
                                value=print_number(count_variants)
                            ),
                            get_mini_container(
                                title="First sample",
                                value=print_date(date_of_first_sample)
                            ),
                            get_mini_container(
                                title="Latest sample",
                                value=print_date(date_of_most_recent_sample)
                            )
                        ]
                    )
                ),
                get_dataset_ena_tab_graphs(queries)
            ]
        ))


def get_dataset_ena_tab_graphs(queries: Queries):
    return html.Div(
        children=[
            html.Br(),
            html.Div(children=[
                html.Div(children=get_plot_library_strategies(queries),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
                html.Div(children=get_plot_number_reads(queries),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
                html.Br(),
                html.Div(children=get_plot_coverage(queries),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
                html.Div(children=get_plot_qualities(queries),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
            ]),
        ])


def get_plot_library_strategies(queries: Queries):
    data = get_data_library_strategies(queries)
    fig = px.bar(data_frame=data, x="library_strategy", y="count", color="library_layout",
                 color_discrete_map=LIBRARY_LAYOUT_COLOR_MAP, barmode="group")
    fig.update_layout(
        margin=MARGIN,
        template=TEMPLATE,
        yaxis={'title': "Num. of samples"},
        xaxis={'title': None},
        legend={'title': None}
    )
    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Library strategies**
        
        The number of samples for each library preparation strategy.
        Only Illumina samples are included, other technologies are excluded from CoVigator at the moment.
        """)
    ]


def get_data_library_strategies(queries):
    sql_query = """
    select count(*) as count, library_strategy, library_layout
    from {table} 
    where finished
    group by library_strategy, library_layout""".format(table=SAMPLE_ENA_TABLE_NAME)
    data = pd.read_sql_query(sql_query, queries.session.bind)
    return data


def get_plot_number_reads(queries: Queries):
    data = get_data_number_reads(queries)

    fig = get_scatter_with_marginals(
        data=data,
        x="num_reads_after",
        y="num_reads_before",
        x_title="Num. of reads after trimming",
        y_title="Reported num. of reads")

    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Number of reads**
       
        The reported number of reads corresponds to the number reported in the ENA API.
        The number of reads after trimming corresponds to the number of reads after running fastp on the FASTQ files.
        No samples are excluded based on the number of reads.
        """)
    ]


def get_data_number_reads(queries):
    sql_query = """
    select s.run_accession as run_accession, j.num_reads as num_reads_after, s.read_count as num_reads_before, s.library_strategy
    from {samples_table} as s join {jobs_table} as j on s.run_accession = j.run_accession 
    where s.finished""".format(samples_table=SAMPLE_ENA_TABLE_NAME, jobs_table=JOB_ENA_TABLE_NAME)
    data = pd.read_sql_query(sql_query, queries.session.bind)
    return data


def get_scatter_with_marginals(data, x, y, x_title, y_title):
    fig = go.FigureWidget()
    for k, v in LIBRARY_STRATEGY_COLOR_MAP.items():
        fig.add_scattergl(
            x=data[data.library_strategy == k][x],
            y=data[data.library_strategy == k][y],
            mode='markers',
            marker=dict(color=v),
            name=k,
            opacity=0.5,
            hovertext=data[data.library_strategy == k].run_accession
        )
        fig.add_box(x=data[data.library_strategy == k][x], yaxis='y2', marker=dict(color=v),
                    showlegend=False)
        fig.add_box(y=data[data.library_strategy == k][y], xaxis='x2', marker=dict(color=v),
                    showlegend=False)
    fig.layout = dict(
        template=TEMPLATE,
        margin=MARGIN,
        yaxis={'title': y_title,
               'domain': [0, 0.85], 'showgrid': True, 'zeroline': False, 'type': 'log'},
        xaxis={'title': x_title,
               'domain': [0, 0.85], 'showgrid': True, 'zeroline': False, 'type': 'log'},
        legend={'title': None, 'itemsizing': 'constant'},
        bargap=0,
        xaxis2={'domain': [0.85, 1], 'showgrid': False, 'zeroline': False, 'visible': False, 'showticklabels': False},
        yaxis2={'domain': [0.85, 1], 'showgrid': False, 'zeroline': False, 'visible': False, 'showticklabels': False},
        hovermode='closest',
    )
    return fig


def get_plot_coverage(queries: Queries):
    data = get_data_coverage(queries)

    fig = get_scatter_with_marginals(
        data=data,
        x="mean_depth",
        y="coverage",
        x_title="Mean depth of coverage",
        y_title="Genome coverage (%)")

    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Horizontal coverage (%)**

        The percentage of the genome covered by at least one read.
        Samples covering less than 20 % of the genome are excluded.
        """)
    ]


def get_data_coverage(queries):
    sql_query = """
    select s.run_accession, s.library_strategy, j.coverage, j.mean_depth
    from {table} as s join {jobs_table} as j on s.run_accession = j.run_accession 
    where finished""".format(table=SAMPLE_ENA_TABLE_NAME, jobs_table=JOB_ENA_TABLE_NAME)
    data = pd.read_sql_query(sql_query, queries.session.bind)
    return data


def get_plot_qualities(queries: Queries):
    data = get_data_qualities(queries)

    fig = get_scatter_with_marginals(
        data=data,
        x="mean_base_quality",
        y="mean_mapping_quality",
        x_title="Mean BCQ",
        y_title="Mean MQ")

    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Mean mapping quality (MQ) and Base Call Quality (BCQ)**
        
        Samples covering with either mapping quality < 10 or base call quality < 10 are excluded
        """)
    ]


def get_data_qualities(queries):
    sql_query = """
    select s.run_accession, s.library_strategy, j.mean_mapping_quality, j.mean_base_quality
    from {table} as s join {jobs_table} as j on s.run_accession = j.run_accession 
    where finished""".format(table=SAMPLE_ENA_TABLE_NAME, jobs_table=JOB_ENA_TABLE_NAME)
    data = pd.read_sql_query(sql_query, queries.session.bind)
    return data


def get_plot_clonal_variants(queries: Queries):
    data = get_data_clonal_variants(queries)
    fig = px.scatter(data_frame=data, color="variant_type", y="vaf", x="dp",
                     log_x=True, log_y=True, opacity=0.5, hover_name="variant_id",
                     marginal_x='box', marginal_y='box',
                     color_discrete_map=VARIANT_TYPE_COLOR_MAP)
    fig.update_layout(
        margin=MARGIN,
        template=TEMPLATE,
        yaxis={'title': "VAF"},
        xaxis={'title': "DP"},
        legend={'title': None}
    )
    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Variant Allele Frequency (VAF) and Depth of Coverage (DP)**

        Variants with VAF >= 0.8 are considered clonal 
        """)
    ]


def get_data_clonal_variants(queries):
    sql_query = """
    select variant_id, vaf, dp, variant_type
    from {table} 
    where source='ENA'""".format(table=VARIANT_OBSERVATION_TABLE_NAME)
    data = pd.read_sql_query(sql_query, queries.session.bind)
    return data


def get_dataset_ena_tab_left_bar(queries: Queries):
    return html.Div(
        className="two columns",
        children=[
            html.Br(),
            dcc.Markdown("""Select one or more genes"""),
            dcc.Dropdown(
                id="sample-dropdown",
                options=[{'label': g.name, 'value': g.name} for g in queries.get_genes()],
                value=None,
                multi=True
            )
        ])



import unittest
from unittest import TestCase

from covigator.configuration import Configuration
from covigator.database.model import SampleEna
from covigator.pipeline.downloader import Downloader, CovigatorMD5CheckSumError
import os

from covigator.tests.unit_tests.faked_objects import FakeConfiguration


class DownloaderTest(TestCase):

    def setUp(self) -> None:
        self.downloader = Downloader(config=FakeConfiguration())

    @unittest.skip
    def test_download_ena_run_with_bad_md5(self):
        run_accession = "TEST12345"
        ena_run = SampleEna(
            run_accession=run_accession,
            fastq_ftp="ftp.sra.ebi.ac.uk/vol1/fastq/ERR419/002/ERR4192722/ERR4192722.fastq.gz;"
                       "ftp.sra.ebi.ac.uk/vol1/fastq/ERR419/002/ERR4192722/ERR4192722_1.fastq.gz;"
                       "ftp.sra.ebi.ac.uk/vol1/fastq/ERR419/002/ERR4192722/ERR4192722_2.fastq.gz",
            fastq_md5="blahblahblah;"
                      "11a6108b741b48402001e0bbae16cbaa;"
                      "2dc39ea2a383550c15013435b65489f8"
        )
        with self.assertRaises(CovigatorMD5CheckSumError):
            self.downloader.download(ena_run)

    def test_download_http(self):
        run_accession = "TEST12346"
        ena_run = SampleEna(
            run_accession=run_accession,
            fastq_ftp="https://tron-mainz.de/wp-content/uploads/2020/07/TRON_Logo_Science.svg",
            fastq_md5="ca72bacad0dfcf665df49bfc53cc8b60"
        )
        self.downloader.download(ena_run)
        run_storage_folder = os.path.join(self.downloader.storage_folder, run_accession)
        self.assertTrue(os.path.exists(run_storage_folder))
        self.assertTrue(os.path.exists(os.path.join(run_storage_folder, "TRON_Logo_Science.svg")))

    def test_download_http_without_md5(self):
        run_accession = "TEST12347"
        with self.assertRaises(AssertionError):
            self.downloader.download(SampleEna(
                run_accession=run_accession,
                fastq_ftp="https://commons.wikimedia.org/wiki/File:Google-Logo.svg",
                fastq_md5=None
            ))
        with self.assertRaises(CovigatorMD5CheckSumError):
            self.downloader.download(SampleEna(
                run_accession=run_accession,
                fastq_ftp="https://commons.wikimedia.org/wiki/File:Google-Logo.svg",
                fastq_md5=""
            ))



from math import sqrt
from typing import List
import numpy as np
import pandas as pd
import plotly
from logzero import logger

from covigator.dashboard.figures import VARIANT_TYPE_COLOR_MAP
from covigator.dashboard.figures.figures import Figures, PLOTLY_CONFIG, MARGIN, TEMPLATE
import plotly.express as px
import plotly.graph_objects as go
import dash_core_components as dcc

from covigator.database.model import DataSource


INDEL_TYPE_COLOR_MAP = {
    "INSERTION_INFRAME": "#ee6002",
    "DELETION_INFRAME": "#09af00",
    "INSERTION_FRAMESHIFT": "#ffddb0",
    "DELETION_FRAMESHIFT": "#defabb",
}


class SampleFigures(Figures):

    def get_annotations_plot(self, data_source: str = None, genes: List[str] = None):
        data = self.queries.get_annotations(data_source=data_source, genes=genes)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            fig = px.bar(data, y="count", x="annotation", log_y=True, color='count',
                         color_continuous_scale=plotly.colors.sequential.Brwnyl)
            #.update_yaxes(categoryorder="total descending")
            fig.update_layout(
                margin=MARGIN,
                template=TEMPLATE,
                legend={'traceorder': 'normal', 'title': None},
                yaxis={'title': "num. variants (log)"},  # , 'autorange': 'reversed'},
                xaxis={'title': None},
            )
            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                **Most common mutation effects**
                
                *Ratio of non synonymous to synonymous SNVs (N/S): {dnds}*
                """.format(dnds=round(data[data.annotation == "missense_variant"]["count"].sum() /
                                data[data.annotation == "synonymous_variant"]["count"].sum(), 3)))
            ]
        return graph

    def get_indels_lengths_plot(self, data_source: str = None, genes: List[str] = None):
        data = self.queries.get_indel_lengths(data_source=data_source, genes=genes)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            fig = px.bar(
                data, y="count", x="length", color="variant_type", color_discrete_map=INDEL_TYPE_COLOR_MAP)
                #.update_yaxes(categoryorder="total descending")
            fig.update_layout(
                margin=MARGIN,
                template=TEMPLATE,
                legend={'traceorder': 'normal', 'title': None},
                yaxis={'title': "num. samples"},  #, 'autorange': 'reversed'},
                xaxis={'title': None},
            )
            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                **Indel length distribution**
                
                *Insertion to deletion ratio: {indel_ratio}*
                """.format(
                    indel_ratio=round(data[
                        data.variant_type.isin(["INSERTION_INFRAME", "INSERTION_FRAMESHIFT"])]["count"].sum() /
                        data[data.variant_type.isin(["DELETION_INFRAME", "DELETION_FRAMESHIFT"])]["count"].sum(), 3)))
            ]
        return graph

    def get_substitutions_plot(self, variant_types: List[str], data_source: str = None, genes: List[str] = None):
        data = self.queries.get_substitutions(data_source=data_source, genes=genes, variant_types=variant_types)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            fig = px.bar(
                data, y="substitution", x="count", color="variant_type", text="rate",
                color_discrete_map=VARIANT_TYPE_COLOR_MAP)\
                .update_yaxes(categoryorder="total descending")
            fig.update_traces(textposition='outside')
            fig.update_layout(
                margin=go.layout.Margin(l=0, r=40, b=0, t=30),  # we need some extra space for some labels overflowing
                template=TEMPLATE,
                legend={'title': None, 'yanchor': "bottom", 'y': 0.01, 'xanchor': "right", 'x': 0.99},
                showlegend=True,
                yaxis={'title': None, 'autorange': 'reversed'},
                xaxis={'title': "num. samples"},
                uniformtext_mode='show',
                uniformtext_minsize=8
            )
            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                **Top mutations**
                
                *Only mutations occurring at least in 10 samples are represented*
                """)
            ]
        return graph

    def get_variants_per_sample_plot(
            self, data_source: str = None, genes: List[str] = None, variant_types: List[str] = None):

        data = self.queries.get_variants_per_sample(data_source=data_source, genes=genes, variant_types=variant_types)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            counts = np.repeat(data.number_mutations, data["count"])
            fig = px.violin(y=counts, orientation='v', box=True, color_discrete_sequence=["grey"])
            fig.update_layout(
                margin=MARGIN,
                template=TEMPLATE,
                legend={'traceorder': 'normal', 'title': None},
                yaxis={'title': "num. mutations"},
                xaxis={'title': None},
            )

            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                        **Mutations per sample**

                        *Median: {median} (IQR: {iqr})*
                        """.format(median=round(np.median(counts), 3),
                                   iqr=round(np.percentile(counts, 75) - np.percentile(counts, 25), 3)))
            ]
        return graph

    def get_accumulated_samples_by_country_plot(self, data_source: DataSource = None, countries=None, min_samples=1000):
        logger.info("start query")
        data = self.queries.get_accumulated_samples_by_country(
            data_source=data_source, countries=countries, min_samples=min_samples)
        logger.info("finished query")
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            countries = list(data.sort_values("cumsum", ascending=False).country.unique())
            fig = px.area(data, x="date", y="cumsum", color="country",
                          category_orders={
                              "country": countries[::-1]},
                          labels={"cumsum": "num. samples", "count": "increment"},
                          hover_data=["count"],
                          color_discrete_sequence=px.colors.qualitative.Light24)
            fig.update_layout(
                margin=MARGIN,
                template=TEMPLATE,
                legend={'traceorder': 'reversed', 'title': None},
                xaxis={'title': None},
            )

            top_countries = countries[0: min(5, len(countries))]
            top_countries_and_cumsum = data[data.country.isin(top_countries)][["country", "cumsum"]] \
                .groupby("country").max().reset_index().sort_values("cumsum", ascending=False)
            top_countries_tooltip = list(
                top_countries_and_cumsum.apply(lambda x: "{} ({})".format(x[0], int(x[1])), axis=1))

            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                **Accumulated samples by country**

                *Top {} countries: {}. Countries with < {} samples are excluded*
                """.format(len(top_countries_tooltip),
                           ", ".join(top_countries_tooltip),
                           min_samples))
            ]
        return graph



import os
import json
from Bio import SeqIO
from sqlalchemy.orm import Session
from covigator.configuration import Configuration
from covigator.database.model import Gene
from logzero import logger


class GeneAnnotationsLoader:

    def __init__(self, session: Session, config: Configuration):
        self.config = config
        self.session = session
        assert self.config.reference_gene_annotations is not None and os.path.exists(self.config.reference_gene_annotations), \
            "Please configure the gene annotations in the variable {}".format(self.config.ENV_COVIGATOR_GENE_ANNOTATIONS)

    def load_data(self):
        # reads the JSON
        with open(self.config.reference_gene_annotations) as fd:
            data = json.load(fd)

        for g in data["genes"]:
            self.session.add(Gene(identifier=g["id"], name=g["name"], start=int(g["start"]), end=int(g["end"]), data=g))

        self.session.commit()
        logger.info("Loaded into the database {} genes".format(len(data["genes"])))



from itertools import cycle
import colorlover
import dash_table
import numpy as np
import re
from covigator.dashboard.figures.figures import Figures, PLOTLY_CONFIG, TEMPLATE, MARGIN
import plotly
import plotly.express as px
import plotly.graph_objects as go
import dash_html_components as html
import dash_core_components as dcc
from covigator.database.model import Gene

VARIANT_TOOLTIP = '<b>%{text}</b><br>' + 'Allele frequency: %{y:.5f}<br>' + 'Genomic Position: %{x}'
GENE_COLORS = cycle(plotly.express.colors.sequential.Reds)
DOMAIN_COLORS = cycle(reversed(plotly.express.colors.sequential.Purples))
OTHER_VARIANT_SYMBOL = "x"
INSERTION_SYMBOL = "triangle-up"
DELETION_SYMBOL = "triangle-down"
MISSENSE_VARIANT_SYMBOL = "circle"
VERY_COMMON_VARIANTS_COLOR = plotly.express.colors.sequential.Reds[-1]
COMMON_VARIANTS_COLOR = plotly.express.colors.sequential.Reds[-3]
RARE_VARIANTS_COLOR = plotly.express.colors.sequential.Reds[-7]
COMMON_VARIANTS_THRESHOLD = 0.1
LOW_FREQUENCY_VARIANTS_THRESHOLD = 0.01
LOW_FREQUENCY_VARIANTS_COLOR = plotly.express.colors.sequential.Reds[-5]
RARE_VARIANTS_THRESHOLD = 0.001
MONTH_PATTERN = re.compile('[0-9]{4}-[0-9]{2}')


class VariantsFigures(Figures):

    def _get_color_by_af(self, af):
        color = None
        if af < RARE_VARIANTS_THRESHOLD:
            color = RARE_VARIANTS_COLOR
        elif RARE_VARIANTS_THRESHOLD <= af < LOW_FREQUENCY_VARIANTS_THRESHOLD:
            color = LOW_FREQUENCY_VARIANTS_COLOR
        elif LOW_FREQUENCY_VARIANTS_THRESHOLD <= af < COMMON_VARIANTS_THRESHOLD:
            color = COMMON_VARIANTS_COLOR
        elif af >= COMMON_VARIANTS_THRESHOLD:
            color = VERY_COMMON_VARIANTS_COLOR
        return color

    def _get_table_style_by_af(self):
        return [
            {
                'if': {
                    'filter_query': '{{frequency}} >= 0 && {{frequency}} < {}'.format(RARE_VARIANTS_THRESHOLD),
                    'column_id': "frequency"
                },
                'backgroundColor': RARE_VARIANTS_COLOR,
                'color': 'inherit'
            },
            {
                'if': {
                    'filter_query': '{{frequency}} >= {} && {{frequency}} < {}'.format(
                        RARE_VARIANTS_THRESHOLD, LOW_FREQUENCY_VARIANTS_THRESHOLD),
                    'column_id': "frequency"
                },
                'backgroundColor': LOW_FREQUENCY_VARIANTS_COLOR,
                'color': 'inherit'
            },
            {
                'if': {
                    'filter_query': '{{frequency}} >= {} && {{frequency}} < {}'.format(
                        LOW_FREQUENCY_VARIANTS_THRESHOLD, COMMON_VARIANTS_THRESHOLD),
                    'column_id': "frequency"
                },
                'backgroundColor': COMMON_VARIANTS_COLOR,
                'color': 'white'
            },
            {
                'if': {
                    'filter_query': '{{frequency}} >= {}'.format(COMMON_VARIANTS_THRESHOLD),
                    'column_id': "frequency"
                },
                'backgroundColor': VERY_COMMON_VARIANTS_COLOR,
                'color': 'white'
            }
        ]

    def get_top_occurring_variants_plot(self, top, gene_name, date_range_start, date_range_end, metric, source):
        data = self.queries.get_top_occurring_variants_precomputed(top, gene_name, metric, source)
        fig = dcc.Markdown("""**No variants for the current selection**""")
        if data is not None and data.shape[0] > 0:
            # removes the columns from the months out of the range
            month_columns = [c for c in data.columns if MONTH_PATTERN.match(c)]
            included_month_colums = [c for c in month_columns if c >= date_range_start and c <= date_range_end]
            excluded_month_colums = [c for c in month_columns if c < date_range_start or c > date_range_end]
            data.drop(excluded_month_colums, axis=1, inplace=True)

            # set the styles of the cells
            styles_counts = self.discrete_background_color_bins(data, columns=included_month_colums)
            styles_total_count = self.discrete_background_color_bins(data, columns=["total"], colors="Reds")
            styles_frequency = self._get_table_style_by_af()
            styles_striped = [{
                'if': {'row_index': 'odd'},
                'backgroundColor': 'rgb(248, 248, 248)'
            }]

            month_columns = [{'name': ["", i], 'id': i} for i in data.columns if i.startswith("20")]
            month_columns[0]['name'][0] = 'Monthly counts' if metric == "count" else 'Monthly frequencies'

            fig = dash_table.DataTable(
                id="top-occurring-variants-table",
                data=data.to_dict('records'),
                sort_action='native',
                columns=[
                            {"name": ["Variant", "Gene"], "id": "gene_name"},
                            {"name": ["", "DNA mutation"], "id": "dna_mutation"},
                            {"name": ["", "Protein mutation"], "id": "hgvs_p"},
                            {"name": ["", "Effect"], "id": "annotation"},
                            {"name": ["", "Frequency"], "id": "frequency"},
                            {"name": ["", "Count"], "id": "total"},
                        ] + month_columns,
                style_data_conditional=styles_striped + styles_counts + styles_frequency + styles_total_count,
                style_cell_conditional=[
                    {
                        'if': {'column_id': c},
                        'textAlign': 'left'
                    } for c in ['gene_name', 'dna_mutation', 'hgvs_p', 'annotation']
                ],
                style_table={'overflowX': 'auto'},
                style_as_list_view=True,
                style_header={
                    'backgroundColor': 'rgb(230, 230, 230)',
                    'fontWeight': 'bold'
                },
                sort_by=[{"column_id": "frequency", "direction": "desc"}],
                row_selectable='multi'
            )

        return [
            fig,
            dcc.Markdown("""
                            **Top occurring variants** 
                            *table shows the {} variants{} with the highest frequency across all samples.*
                            *The counts and frequencies per month are only shown between {} and {}.*
                            *Selections in this table will be highlighted in the genome view and in the co-occurrence matrix.*
                            """.format(top, " in gene {}".format(gene_name) if gene_name else "",
                                       date_range_start, date_range_end))
        ]

    def discrete_background_color_bins(self, df, n_bins=5, columns='all', colors='Blues'):
        bounds = [i * (1.0 / n_bins) for i in range(n_bins + 1)]
        if columns == 'all':
            if 'id' in df:
                df_numeric_columns = df.select_dtypes('number').drop(['id'], axis=1)
            else:
                df_numeric_columns = df.select_dtypes('number')
        else:
            df_numeric_columns = df[columns]
        df_max = df_numeric_columns.max().max()
        df_min = df_numeric_columns.min().min()
        ranges = [
            ((df_max - df_min) * i) + df_min
            for i in bounds
        ]
        styles = []
        for i in range(1, len(bounds)):
            min_bound = ranges[i - 1]
            max_bound = ranges[i]
            backgroundColor = colorlover.scales[str(n_bins)]['seq'][colors][i - 1]
            color = 'white' if i > len(bounds) / 2. else 'inherit'

            for column in df_numeric_columns:
                styles.append({
                    'if': {
                        'filter_query': (
                                '{{{column}}} >= {min_bound}' +
                                (' && {{{column}}} < {max_bound}' if (i < len(bounds) - 1) else '')
                        ).format(column=column, min_bound=min_bound, max_bound=max_bound),
                        'column_id': column
                    },
                    'backgroundColor': backgroundColor,
                    'color': color
                })

        return styles

    def get_cooccurrence_heatmap(self, gene_name, selected_variants, metric="jaccard", min_occurrences=5):
        data = self.queries.get_variants_cooccurrence_by_gene(gene_name=gene_name, min_cooccurrence=min_occurrences)
        graph = dcc.Markdown("""**No co-occurrent variants for the current selection**""")
        if data is not None and data.shape[0] > 0:

            all_variants = data.variant_id_one.unique()
            values = np.array_split(data[metric], len(all_variants))
            texts = np.array_split(data.hgvs_tooltip, len(all_variants))
            if metric == "count":
                hovertemplate = '<b>%{text}</b><br>' + 'Counts: %{z}<br>' + 'Variant one: %{x}<br>' + 'Variant two: %{y}'
            elif metric == "frequency":
                hovertemplate = '<b>%{text}</b><br>' + 'Frequency: %{z:.3f}<br>' + 'Variant one: %{x}<br>' + 'Variant two: %{y}'
            elif metric == "jaccard":
                hovertemplate = '<b>%{text}</b><br>' + 'Jaccard index: %{z:.3f}<br>' + 'Variant one: %{x}<br>' + 'Variant two: %{y}'
            elif metric == "kappa":
                hovertemplate = '<b>%{text}</b><br>' + 'Kappa coefficient: %{z:.3f}<br>' + 'Variant one: %{x}<br>' + 'Variant two: %{y}'
            heatmap = go.Heatmap(
                z=values,
                x=all_variants,
                y=all_variants,
                colorscale="Oryel",
                hoverongaps=False,
                text=texts,
                hovertemplate=hovertemplate,
            )
            if selected_variants:
                selected_variant_ids = [v.get("dna_mutation") for v in selected_variants]
                selected_data = data
                selected_data[metric].where(
                    (selected_data.variant_id_one.isin(selected_variant_ids)) |
                    (selected_data.variant_id_two.isin(selected_variant_ids)), np.nan, inplace=True)
                values_selected = np.array_split(selected_data[metric], len(all_variants))
                texts_selected = np.array_split(selected_data.hgvs_tooltip, len(all_variants))

                heatmap_selected = go.Heatmap(
                    z=values_selected,
                    x=all_variants,
                    y=all_variants,
                    colorscale="Teal",
                    hoverongaps=False,
                    showscale=False,
                    text=texts_selected,
                    hovertemplate=hovertemplate,
                )
            layout = go.Layout(
                template=TEMPLATE,
                margin=MARGIN,
                height=700,
                yaxis=dict(visible=True, tickfont={"size": 10}, showgrid=False, showspikes=True, spikemode='toaxis',
                           spikethickness=2),
                xaxis=dict(tickangle=-45, tickfont={"size": 10}, showgrid=False, showspikes=True, spikemode='toaxis',
                           spikethickness=2),
            )
            traces = [heatmap]
            if selected_variants:
                traces.append(heatmap_selected)
            fig = go.Figure(data=traces, layout=layout)

            # the y index is reversed in plotly heatmap
            fig.update_yaxes(autorange="reversed")
            graph = dcc.Graph(figure=fig, config=PLOTLY_CONFIG)

        return [
            graph,
            dcc.Markdown("""
                        ***Co-occurrence matrix*** *showing variant pairs co-occurring in at least {} samples (this value is configurable).*
                        *The metric in the co-occurrence matrix can be chosen among counts, frequencies, Jaccard index or 
                        Cohen's kappa coefficient. The Cohen's kappa coefficient introduces a correction to the Jaccard index for
                        variants with low occurrence.*
                        *The diagonal contains the total counts or just 1.0 in the other metrics.*
                        *The upper diagonal is not shown for clarity.*
                        *Synonymous variants are excluded.*
                        *Different genomic variants causing the same protein variant are not grouped.*
                        """.format(min_occurrences, metric, " on gene {}".format(gene_name) if gene_name else ""))
        ]

    def get_variants_abundance_plot(self, bin_size=50, source=None):

        # reads genes and domains across the whole genome
        genes = self.queries.get_genes()
        domains = []
        for g in genes:
            domains.extend([(g, d) for d in g.get_pfam_domains()])

        # reads variants abundance
        variant_abundance = self.queries.get_variant_abundance_histogram(bin_size=bin_size, source=source)

        # reads conservation and bins it
        conservation = self.queries.get_conservation_table(bin_size=bin_size)

        # joins variant abundance and conservation
        data = variant_abundance.set_index("position_bin").join(conservation.set_index("position_bin"))
        data.reset_index(inplace=True)
        data.fillna(0, inplace=True)

        layout = go.Layout(
            template=TEMPLATE,
            margin=MARGIN,
            xaxis=dict(domain=[0, 1.0], tickformat=',d', hoverformat=',d', ticksuffix=" bp", ticks="outside",
                       visible=True, anchor="y7", showspikes=True, spikemode='across', spikethickness=2),
            yaxis=dict(domain=[0.9, 1.0], anchor='x7'),
            yaxis2=dict(domain=[0.6, 0.9], anchor='x7'),
            yaxis3=dict(domain=[0.45, 0.6], anchor='x7', visible=False),
            yaxis4=dict(domain=[0.3, 0.45], anchor='x7', visible=False),
            yaxis5=dict(domain=[0.15, 0.3], anchor='x7', visible=False),
            yaxis6=dict(domain=[0.05, 0.1], anchor='x7', visible=False),
            yaxis7=dict(domain=[0.0, 0.05], anchor='x7', visible=False),
            legend={'traceorder': 'normal'}
        )

        gene_traces = [self._get_gene_trace(g, color=c, yaxis='y6') for g, c in zip(genes, GENE_COLORS)]
        domain_traces = [self._get_domain_trace(color=c, domain=d, gene=g, yaxis='y7')
                         for (g, d), c in zip(domains, DOMAIN_COLORS)]
        conservation_traces = self._get_conservation_traces(
            conservation, xaxis='x', yaxis1='y3', yaxis2='y4', yaxis3='y5')
        variant_counts_traces = [
            go.Scatter(x=data.position_bin, y=data.count_variant_observations,
                       name="All variants", text="All variants", showlegend=False,
                       line_color=plotly.express.colors.sequential.Blues[-2], line_width=1),
            go.Scatter(x=data.position_bin,
                       y=[data.count_unique_variants.mean() for _ in range(data.shape[0])],
                       yaxis='y2', name="Mean unique variants", text="Mean unique variants",
                       line_width=1,
                       showlegend=False, line_color=plotly.express.colors.sequential.Blues[-3]),
            go.Scatter(x=data.position_bin, y=data.count_unique_variants, yaxis='y2',
                       name="Unique variants", text="Unique variants", showlegend=False, fill='tonexty',
                       line_color=plotly.express.colors.sequential.Blues[-4], line_width=1)
        ]

        fig = go.Figure(data=variant_counts_traces + conservation_traces + gene_traces + domain_traces, layout=layout)

        # add track names
        fig.add_annotation(x=0.98, y=1.1, xref="x domain", yref="y domain", text="All variants",
                           showarrow=False, yshift=10)
        fig.add_annotation(x=0.98, y=0.9, xref="x domain", yref="y2 domain", text="Unique variants",
                           showarrow=False, yshift=10)
        fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y3 domain", text="Conservation SARS-CoV-2",
                           showarrow=False, yshift=10)
        fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y4 domain", text="Conservation SARS-like betaCoV",
                           showarrow=False, yshift=10)
        fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y5 domain", text="Conservation vertebrate CoV",
                           showarrow=False, yshift=10)

        return [
            dcc.Graph(
                figure=fig,
                config=PLOTLY_CONFIG
            ),
            dcc.Markdown("""
                ***Genome view*** *representing the abundance of variants and ConsHMM (Arneson, 2019) conservation 
                using a bin size of {} bp. Synonymous variants are included.*

                *The first track shows the count of variants across the genome including repetitions. *
                *The second track shows the count of unique variants across the genome, the horizontal line represents 
                the average number of unique variants per bin and thus distinguishes regions with over and under the 
                average number of unique variants.*
                *The third, fourth and fifth tracks represent the conservation as reported by ConsHMM within 
                SARS-CoV-2, among SARS-like betaCoV and among vertebrate CoV. Correlation between distribution of 
                unique variants and conservation within Sars-CoV-2, among SARS-like betacoronavirus and among 
                vertebrates CoV respectively: {}, {}, {}.*
                *Genes and Pfam domains are represented in tones of red and purple respectively.*

                *Conservation data source: https://github.com/ernstlab/ConsHMM_CoV*

                *Arneson A, Ernst J. Systematic discovery of conservation states for single-nucleotide annotation of the 
                human genome. Communications Biology, 248, 2019. doi: https://doi.org/10.1038/s42003-019-0488-1*
                """.format(bin_size,
                           round(np.corrcoef(data.conservation, data.count_unique_variants)[0][1], 5),
                           round(np.corrcoef(data.conservation_sarbecovirus, data.count_unique_variants)[0][1], 5),
                           round(np.corrcoef(data.conservation_vertebrates, data.count_unique_variants)[0][1], 5)
                           ))]

    def get_variants_plot(self, gene_name, selected_variants, bin_size, source=None):

        # reads gene annotations
        gene = self.queries.get_gene(gene_name)
        pfam_protein_features = gene.get_pfam_domains()

        # reads variants
        variants = self.queries.get_non_synonymous_variants_by_region(start=gene.start, end=gene.end, source=source)

        # reads conservation and bins it
        conservation = self.queries.get_conservation_table(start=gene.start, end=gene.end, bin_size=bin_size)

        if variants.shape[0] > 0:
            # reads total number of samples and calculates frequencies
            count_samples = self.queries.count_samples(source=source)
            variants["af"] = variants.count_occurrences / count_samples
            variants["log_af"] = variants.af.transform(lambda x: np.log(x + 1))
            variants["log_count"] = variants.count_occurrences.transform(lambda x: np.log(x))
            # TODO: do something in the data ingestion about multiple annotations on the same variant
            variants.annotation = variants.annotation.transform(lambda a: a.split("&")[0])

            main_xaxis = 'x'

            variants_traces = []
            missense_variants = variants[variants.annotation == "missense_variant"]
            if missense_variants.shape[0] > 0:
                variants_traces.append(self._get_variants_scatter(
                    missense_variants, name="missense variants", symbol=MISSENSE_VARIANT_SYMBOL, xaxis=main_xaxis))

            deletion_variants = variants[variants.annotation.isin(
                ["disruptive_inframe_deletion", "conservative_inframe_deletion"])]
            if deletion_variants.shape[0] > 0:
                variants_traces.append(self._get_variants_scatter(
                    deletion_variants, name="inframe deletions", symbol=DELETION_SYMBOL, xaxis=main_xaxis))

            insertion_variants = variants[variants.annotation.isin(
                ["disruptive_inframe_insertion", "conservative_inframe_insertion"])]
            if insertion_variants.shape[0] > 0:
                variants_traces.append(self._get_variants_scatter(
                    insertion_variants, name="inframe insertions", symbol=INSERTION_SYMBOL, xaxis=main_xaxis))

            other_variants = variants[~variants.annotation.isin([
                "missense_variant", "disruptive_inframe_deletion", "conservative_inframe_deletion",
                "disruptive_inframe_insertion", "conservative_inframe_insertion"])]
            if other_variants.shape[0] > 0:
                variants_traces.append(self._get_variants_scatter(
                    other_variants, name="other variants", symbol=OTHER_VARIANT_SYMBOL, xaxis=main_xaxis))

            selected_variants_trace = None
            if selected_variants:
                selected_variants_trace = go.Scatter(
                    x=[int(v.get("dna_mutation").split(":")[0]) for v in selected_variants],
                    y=[v.get("frequency") for v in selected_variants],
                    name="selected variants",
                    mode='markers',
                    marker=dict(
                        symbol="circle",
                        color="blue",
                        size=10,
                        showscale=False
                    ),
                    xaxis=main_xaxis,
                    showlegend=True,
                    text=["{} ({})".format(v.get("hgvs_p"), v.get("annotation")) for v in selected_variants],
                    hovertemplate=VARIANT_TOOLTIP
                )

            gene_trace = self._get_gene_trace(
                gene, color=plotly.express.colors.sequential.Reds[1], yaxis='y5', xaxis=main_xaxis)
            domain_traces = [self._get_domain_trace(
                color=c, gene=gene, domain=d, yaxis='y6', xaxis=main_xaxis, showlegend=True)
                for d, c in zip(pfam_protein_features, DOMAIN_COLORS)]
            conservation_traces = self._get_conservation_traces(
                conservation, main_xaxis, yaxis1='y2', yaxis2='y3', yaxis3='y4')

            data = variants_traces + conservation_traces + [gene_trace] + domain_traces
            if selected_variants_trace is not None:
                data.append(selected_variants_trace)

            layout = go.Layout(
                template=TEMPLATE,
                margin=go.layout.Margin(l=0, r=0, b=0, t=20),
                xaxis=dict(tickformat=',d', hoverformat=',d', ticksuffix=" bp", ticks="outside",
                           showspikes=True, spikemode='across', spikethickness=1, anchor='y6'),
                yaxis=dict(title='Allele frequency', type='log', domain=[0.4, 1.0], anchor=main_xaxis),
                yaxis2=dict(domain=[0.3, 0.4], visible=False, anchor=main_xaxis),
                yaxis3=dict(domain=[0.2, 0.3], visible=False, anchor=main_xaxis),
                yaxis4=dict(domain=[0.1, 0.2], visible=False, anchor=main_xaxis),
                yaxis5=dict(domain=[0.05, 0.1], visible=False, anchor=main_xaxis),
                yaxis6=dict(domain=[0.0, 0.05], visible=False, anchor=main_xaxis)
            )

            fig = go.Figure(data=data, layout=layout)

            fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y2 domain", text="Conservation SARS-CoV-2",
                               showarrow=False, yshift=10)
            fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y3 domain", text="Conservation SARS-like betaCoV",
                               showarrow=False, yshift=10)
            fig.add_annotation(x=0.98, y=1.0, xref="x domain", yref="y4 domain", text="Conservation vertebrate CoV",
                               showarrow=False, yshift=10)

            # add horizontal lines on the frequency boundaries
            fig.add_hline(y=0.1, line_width=1, line_dash="dash", line_color=VERY_COMMON_VARIANTS_COLOR)
            fig.add_hline(y=0.01, line_width=1, line_dash="dash", line_color=COMMON_VARIANTS_COLOR)
            fig.add_hline(y=0.001, line_width=1, line_dash="dash", line_color=RARE_VARIANTS_COLOR)

            return [dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                    dcc.Markdown("""
                        ***Gene view*** *representing each variant with its frequency in the population and 
                        ConsHMM (Arneson, 2019) conservation using a bin size of {} bp. Synonymous variants and variants 
                        occurring in a single sample are excluded.*

                        *The scatter plot shows non synonymous variants occurring in at least two samples on gene {}. 
                        The x-axis shows the genomic coordinates and the y-axis shows the allele frequency.*
                        *The category of "other variants" includes frameshift indels, stop codon gain and lost and 
                        start lost variants.*
                        *The variants are colored according to their frequency as rare variants (< 0.1 %), 
                        low frequency variants (>= 0.1% and < 1%), common variants (>= 1% and < 10%) and very common 
                        variants (>= 10%). The second, third and fourth tracks in grey represent the conservation as 
                        reported by ConsHMM within SARS-CoV-2, among SARS-like betaCoV and among vertebrate CoV.*
                        *Genes and Pfam domains are represented in tones of red and purple respectively.*

                        *Conservation data source: https://github.com/ernstlab/ConsHMM_CoV*

                        *Arneson A, Ernst J. Systematic discovery of conservation states for single-nucleotide annotation of the 
                        human genome. Communications Biology, 248, 2019. doi: https://doi.org/10.1038/s42003-019-0488-1*
                        """.format(bin_size, gene_name))]
        else:
            return dcc.Markdown("""**No variants for the current selection**""")

    def _get_conservation_traces(self, conservation, xaxis, yaxis1, yaxis2, yaxis3):
        return [
            go.Scatter(x=conservation.position_bin, y=conservation.conservation, yaxis=yaxis1, xaxis=xaxis,
                       text="Conservation SARS-CoV-2", textposition="top right", showlegend=False,
                       fill='tozeroy', line_color="grey", line_width=1),
            go.Scatter(x=conservation.position_bin, y=conservation.conservation_sarbecovirus, yaxis=yaxis2,
                       xaxis=xaxis, text="Conservation SARS-like betacoronavirus", textposition="top right",
                       showlegend=False, fill='tozeroy', line_color="grey", line_width=1),
            go.Scatter(x=conservation.position_bin, y=conservation.conservation_vertebrates, yaxis=yaxis3,
                       xaxis=xaxis, text="Conservation vertebrates", textposition="top right",
                       showlegend=False, fill='tozeroy', line_color="grey", line_width=1)
        ]

    @staticmethod
    def _get_domain_trace(color: str, domain: dict, gene: Gene, yaxis='y', xaxis='x', showlegend=False):
        domain_start = gene.start + int(domain["start"])
        domain_end = gene.start + int(domain["end"])
        domain_name = domain.get('description')
        return go.Scatter(
            mode='lines',
            x=[domain_start, domain_end, domain_end, domain_start],
            y=[0, 0, 1, 1],
            name=domain_name,
            fill="toself",
            fillcolor=color,
            text="<b>{} {}</b>: {}-{}".format(gene.name, domain_name, domain_start, domain_end),
            hoverinfo="text",
            line=dict(width=0),
            yaxis=yaxis,
            xaxis=xaxis,
            legendgroup='domains',
            showlegend=showlegend
        )

    @staticmethod
    def _get_gene_trace(gene: Gene, color, yaxis="y", xaxis='x'):
        return go.Scatter(
            mode='lines',
            x=[gene.start, gene.end, gene.end, gene.start],
            y=[0, 0, 1, 1],
            name=gene.name,
            fill="toself",
            fillcolor=color,
            text="<b>{}</b>: {}-{}".format(gene.name, gene.start, gene.end),
            hoverinfo="text",
            line=dict(width=0),
            yaxis=yaxis,
            xaxis=xaxis,
            legendgroup='gene'
        )

    def _get_variants_scatter(self, variants, name, symbol, xaxis='x'):
        return go.Scatter(
            x=variants.position,
            y=variants.af,
            name=name,
            mode='markers',
            # opacity=0.5,
            marker=dict(
                symbol=symbol,
                color=variants.af.transform(self._get_color_by_af),
                showscale=False
            ),
            xaxis=xaxis,
            showlegend=True,
            text=variants[["hgvs_p", "annotation"]].apply(lambda x: "{} ({})".format(x[0], x[1]), axis=1),
            hovertemplate=VARIANT_TOOLTIP
        )

    def get_variants_clustering(self, gene_name, selected_variants, min_cooccurrence, min_samples):
        data = self.queries.get_mds(
            gene_name=gene_name, min_cooccurrence=min_cooccurrence, min_samples=min_samples)

        tables = []
        for c in data.cluster.unique():
            tables.append(dash_table.DataTable(
                id="cluster{}-variants-table".format(c),
                data=data[data.cluster == c].to_dict('records'),
                columns=[
                    {"name": [
                        "Cluster {} (mean Jaccard={})".format(c, data[data.cluster == c].cluster_jaccard_mean.iloc[0]),
                        "DNA mutation"], "id": "variant_id"},
                    {"name": ["", "Protein mutation"], "id": "hgvs_p"},
                ],
                fixed_columns={'headers': True, 'data': 1},
                style_table={'overflowX': 'auto'},
                style_cell={'minWidth': '50px', 'width': '50px', 'maxWidth': '50px'},
                style_as_list_view=True,
                style_header={
                    'backgroundColor': 'rgb(230, 230, 230)',
                    'fontWeight': 'bold'
                },
                sort_by=[{"column_id": "variant_id", "direction": "asc"}],
            ))
            tables.append(html.Br())

        return [
            html.Div(children=tables),
            dcc.Markdown("""
            ***Co-occurrence clustering*** *shows the resulting clusters from the
            co-occurrence matrix with the Jaccard index corrected with the Cohen's kappa coefficient. 
            The co-occurrence matrix is built taking into account only variants with at least {} pairwise 
            co-occurrences and if a gene is provided only variants within that gene.  
            Clustering is performed on the co-occurrence matrix using OPTICS. 
            The mimimum number of neighbours to call
            a cluster is {}.
            Variants selected in the top occurrent variants table are highlighted with a greater size in the plot.
            *

            *Ankerst et al. “OPTICS: ordering points to identify the clustering structure.” ACM SIGMOD Record 28, no. 2 (1999): 49-60.*
            """.format(min_cooccurrence, min_samples))]



import unittest
from itertools import combinations
from unittest import TestCase

from covigator.database.precomputed import Precomputer
from faker import Faker
import numpy as np
from covigator.database.database import Database
from covigator.database.model import JobStatus, DataSource, Sample, Gene
from covigator.database.queries import Queries
from covigator.tests.unit_tests.faked_objects import FakeConfiguration
from covigator.tests.unit_tests.mocked import get_mocked_ena_sample, get_mocked_log, get_mocked_variant, \
    get_mocked_variant_cooccurrence, get_mocked_variant_observation


class QueriesTests(TestCase):

    def setUp(self) -> None:
        # intialise database
        self.database = Database(test=True, verbose=True, config=FakeConfiguration())
        self.session = self.database.get_database_session()
        self.queries = Queries(session=self.session)
        self.faker = Faker()

    def test_get_date_of_first_ena_sample(self):
        first_sample_date = None
        # adds 50 loaded and 50 failed samples, the date should only take into account loaded samples
        samples = [get_mocked_ena_sample(faker=self.faker) for _ in range(50)] + \
                  [get_mocked_ena_sample(faker=self.faker, job_status=JobStatus.FAILED_LOAD) for _ in range(50)]
        for sample_ena, sample, job in samples:
            self.session.add_all([sample_ena, sample, job])
            if job.status == JobStatus.FINISHED:
                if first_sample_date is None:
                    first_sample_date = sample_ena.collection_date
                if sample_ena.collection_date < first_sample_date:
                    first_sample_date = sample_ena.collection_date
        self.session.commit()
        observed_date = self.queries.get_date_of_first_sample()
        self.assertEqual(observed_date, first_sample_date.date())

    def test_get_date_of_first_ena_sample_empty(self):
        observed_date = self.queries.get_date_of_first_sample()
        self.assertIsNone(observed_date)

    def test_get_date_of_most_recent_ena_sample(self):
        most_recent_sample_date = None
        # adds 50 loaded and 50 failed samples, the date should only take into account loaded samples
        samples = [get_mocked_ena_sample(faker=self.faker) for _ in range(50)] + \
                  [get_mocked_ena_sample(faker=self.faker, job_status=JobStatus.FAILED_LOAD) for _ in range(50)]
        for sample_ena, sample, job in samples:
            self.session.add_all([sample_ena, sample, job])
            if job.status == JobStatus.FINISHED:
                if most_recent_sample_date is None:
                    most_recent_sample_date = sample_ena.collection_date
                if sample_ena.collection_date > most_recent_sample_date:
                    most_recent_sample_date = sample_ena.collection_date
        self.session.commit()
        observed_date = self.queries.get_date_of_most_recent_sample()
        self.assertEqual(observed_date, most_recent_sample_date.date())

    def test_get_date_of_most_recent_ena_sample_empty(self):
        observed_date = self.queries.get_date_of_most_recent_sample()
        self.assertIsNone(observed_date)

    def test_get_date_of_last_ena_check(self):
        logs = [get_mocked_log(faker=self.faker, source=DataSource.ENA) for _ in range(25)] + \
                [get_mocked_log(faker=self.faker, source=DataSource.GISAID) for _ in range(25)]
        self.session.add_all(logs)
        self.session.commit()
        observed_date = self.queries.get_date_of_last_check(data_source=DataSource.ENA)
        self.assertIsNotNone(observed_date)
        observed_date_gisaid = self.queries.get_date_of_last_check(data_source=DataSource.GISAID)
        self.assertIsNotNone(observed_date_gisaid)

    def test_get_date_of_last_ena_check_empty(self):
        observed_date = self.queries.get_date_of_last_check(data_source=DataSource.ENA)
        self.assertIsNone(observed_date)
        observed_date = self.queries.get_date_of_last_check(data_source=DataSource.GISAID)
        self.assertIsNone(observed_date)

    def test_get_date_of_last_ena_update(self):
        # NOTE: the implementation that works in Postgres does not work in SQLite!!
        logs = [get_mocked_log(faker=self.faker, source=DataSource.ENA) for _ in range(50)]
        self.session.add_all(logs)
        self.session.commit()
        observed_date = self.queries.get_date_of_last_update(data_source=DataSource.ENA)
        self.assertIsNotNone(observed_date)

    def test_get_date_of_last_ena_update_empty(self):
        observed_date = self.queries.get_date_of_last_update(data_source=DataSource.ENA)
        self.assertIsNone(observed_date)
        observed_date = self.queries.get_date_of_last_update(data_source=DataSource.GISAID)
        self.assertIsNone(observed_date)

    def test_get_cooccurrence_matrix_by_gene_no_data(self):
        Precomputer(session=self.session).load_table_counts()
        data = self.queries.get_variants_cooccurrence_by_gene(gene_name="S", test=True)
        self.assertIsNone(data)

    def test_get_cooccurrence_matrix_by_gene(self):
        other_variants, variants = self._mock_cooccurrence_matrix()
        Precomputer(session=self.session).load_table_counts()

        data = self.queries.get_variants_cooccurrence_by_gene(gene_name="S", min_cooccurrence=1, test=True)
        self.assertIsNotNone(data)
        num_unique_variants = len(set([v.variant_id for v in variants]))
        self.assertEqual(data.shape[0], num_unique_variants * num_unique_variants)
        self.assertEqual(data.shape[1], 7)
        self.assertGreaterEqual(data[data["count"] > 0].shape[0], len(variants))
        self.assertGreaterEqual(data[data["frequency"] > 0].shape[0], len(variants))
        self.assertGreaterEqual(data[data["jaccard"] > 0].shape[0], len(variants))

        data = self.queries.get_variants_cooccurrence_by_gene(gene_name="S", min_cooccurrence=11, test=True)
        self.assertIsNone(data)

        data = self.queries.get_variants_cooccurrence_by_gene(gene_name="X", min_cooccurrence=1, test=True)
        self.assertIsNone(data)

        data = self.queries.get_variants_cooccurrence_by_gene(gene_name="N", min_cooccurrence=1, test=True)
        self.assertIsNotNone(data)
        self.assertEqual(data.shape[1], 7)
        self.assertGreaterEqual(data[data["count"] > 0].shape[0], len(other_variants))
        self.assertGreaterEqual(data[data["frequency"] > 0].shape[0], len(variants))
        self.assertGreaterEqual(data[data["jaccard"] > 0].shape[0], len(variants))

    @unittest.skip
    def test_get_mds(self):
        self._mock_cooccurrence_matrix()

        mds_fit, mds_coords = self.queries.get_mds(gene_name="S")
        self.assertIsNotNone(mds_fit)

    def _mock_cooccurrence_matrix(self):
        # add some variants belonging to two genes
        chromosome = "fixed_chromosome"
        gene_name = "S"
        variants = [get_mocked_variant(faker=self.faker, chromosome=chromosome, gene_name=gene_name) for _ in range(5)]
        other_gene_name = "N"
        other_variants = [get_mocked_variant(faker=self.faker, chromosome=chromosome, gene_name=other_gene_name) for _
                          in range(5)]
        self.session.add_all(variants + other_variants)
        self.session.commit()
        # adds some cooccurrences
        cooccurrences = []
        other_cooccurrences = []
        variants_to_sample = {"{}-{}".format(v1.hgvs_p, v2.hgvs_p): (v1, v2) for v1, v2 in
                              list(combinations(variants, 2))}
        other_variants_to_sample = {"{}-{}".format(v1.hgvs_p, v2.hgvs_p): (v1, v2) for v1, v2 in
                                    list(combinations(other_variants, 2))}
        combined_variants_to_sample = {"{}-{}".format(v1.hgvs_p, v2.hgvs_p): (v1, v2) for v1, v2 in
                                       list(zip(variants, other_variants))}
        for variant in variants + other_variants:
            cooccurrences.append(get_mocked_variant_cooccurrence(self.faker, variant, variant))
        for (variant_one, variant_two) in [variants_to_sample.get(k) for k in
                                           np.random.choice(list(variants_to_sample.keys()), 5, replace=False)]:
            cooccurrences.append(get_mocked_variant_cooccurrence(self.faker, variant_one, variant_two))
        for (variant_one, variant_two) in [other_variants_to_sample.get(k) for k in
                                           np.random.choice(list(other_variants_to_sample.keys()), 5, replace=False)]:
            other_cooccurrences.append(get_mocked_variant_cooccurrence(self.faker, variant_one, variant_two))
        for (variant_one, variant_two) in [combined_variants_to_sample.get(k) for k in
                                           np.random.choice(list(combined_variants_to_sample.keys()), 5,
                                                            replace=False)]:
            other_cooccurrences.append(get_mocked_variant_cooccurrence(self.faker, variant_one, variant_two))
        self.session.add_all(cooccurrences + other_cooccurrences)
        self.session.commit()
        # add some samples to compute the frequency right
        for _ in range(10):
            sample_ena, sample, job = get_mocked_ena_sample(Faker())
            self.session.add(sample_ena)
            self.session.add(sample)
            self.session.add(job)
            self.session.commit()
        return other_variants, variants

    def test_get_variant_abundance_histogram(self):

        # gets an empty histogram
        histogram = self.queries.get_variant_abundance_histogram(cache=False)
        self.assertIsNone(histogram)

        # mocks a 100 variants
        num_variants = 100
        variants = [get_mocked_variant(faker=self.faker, chromosome="chr_test") for _ in range(num_variants)]
        self.session.add_all(variants)
        self.session.commit()

        # gets an histogram over 100 variants without variant observations
        histogram = self.queries.get_variant_abundance_histogram(cache=False)
        self.assertIsNotNone(histogram)
        self.assertGreater(histogram.shape[0], 0)
        self.assertEqual(histogram.shape[1], 3)
        self.assertEqual(histogram.count_unique_variants.sum(), num_variants)
        self.assertEqual(histogram.count_variant_observations.sum(), 0)

        # mock some variant observations
        test_samples = [Sample(id=self.faker.unique.uuid4(), source=DataSource.ENA) for _ in range(5)]
        self.session.add_all(test_samples)
        self.session.commit()
        variant_observations = []
        for v in variants:
            for i in range(self.faker.random_int(min=1, max=5)):
                variant_observations.append(get_mocked_variant_observation(variant=v, sample=test_samples[i]))
        self.session.add_all(variant_observations)
        self.session.commit()

        # gets an histogram over 100 variants
        histogram = self.queries.get_variant_abundance_histogram(cache=False)
        self.assertIsNotNone(histogram)
        self.assertGreater(histogram.shape[0], 0)
        self.assertEqual(histogram.shape[1], 3)
        self.assertEqual(histogram.count_unique_variants.sum(), num_variants)
        self.assertEqual(histogram.count_variant_observations.sum(), len(variant_observations))

    def test_get_conservation_table(self):
        conservation50 = self.queries.get_conservation_table(bin_size=50)
        self.assertIsNotNone(conservation50)
        self.assertEqual(conservation50.shape[1], 4)
        self.assertGreater(conservation50.shape[0], 0)
        self.assertEqual(conservation50[conservation50.conservation.isna()].shape[0], 0)
        self.assertEqual(conservation50[conservation50.conservation_sarbecovirus.isna()].shape[0], 0)
        self.assertEqual(conservation50[conservation50.conservation_vertebrates.isna()].shape[0], 0)

        conservation10 = self.queries.get_conservation_table(bin_size=10)
        self.assertGreater(conservation10.shape[0], conservation50.shape[0])
        self.assertEqual(conservation10.shape[1], conservation50.shape[1])

        conservation5 = self.queries.get_conservation_table(bin_size=5)
        self.assertGreater(conservation5.shape[0], conservation50.shape[0])
        self.assertGreater(conservation5.shape[0], conservation10.shape[0])
        self.assertEqual(conservation5.shape[1], conservation50.shape[1])

        conservation5_smaller = self.queries.get_conservation_table(bin_size=5, start=10000, end=20000)
        self.assertGreater(conservation5.shape[0], conservation5_smaller.shape[0])
        self.assertEqual(conservation5.shape[1], conservation50.shape[1])

    def test_get_genes_metadata(self):
        genes = self.queries.get_genes()
        self.assertIsNotNone(genes)
        self.assertGreater(len(genes), 0)
        for g in genes:
            self.assertIsInstance(g, Gene)

    def test_get_genes(self):
        genes = self.queries.get_genes()
        self.assertIsNotNone(genes)
        self.assertGreater(len(genes), 0)
        for g in genes:
            self.assertIsInstance(g, Gene)

    def test_get_gene(self):
        gene = self.queries.get_gene("S")
        self.assertIsNotNone(gene)
        self.assertIsInstance(gene, Gene)
        gene = self.queries.get_gene("NOEXISTO")
        self.assertIsNone(gene)

    def test_count_jobs_in_queue(self):
        samples = [get_mocked_ena_sample(faker=self.faker, job_status=JobStatus.QUEUED) for _ in range(50)] + \
                  [get_mocked_ena_sample(faker=self.faker, job_status=JobStatus.FINISHED) for _ in range(50)]
        for sample_ena, sample, job in samples:
            self.session.add_all([sample_ena, sample, job])
        self.session.commit()
        count_jobs_in_queue = self.queries.count_jobs_in_queue(DataSource.ENA)
        self.assertEqual(count_jobs_in_queue, 50)
        count_jobs_in_queue = self.queries.count_jobs_in_queue(DataSource.GISAID)
        self.assertEqual(count_jobs_in_queue, 0)



from itertools import combinations
from sqlalchemy.orm import Session
from covigator.database.model import Sample, VariantCooccurrence
from logzero import logger
from covigator.database.queries import Queries
from sqlalchemy.exc import IntegrityError


class CooccurrenceMatrixException(Exception):
    pass


class CooccurrenceMatrix:

    def compute(self, sample: Sample, session: Session):

        assert sample is not None, "Missing sample"
        assert sample.id is not None or sample.id == "", "Missing sample identifier"
        assert session is not None, "Missing DB session"

        queries = Queries(session=session)
        sample_id = sample.id
        logger.info("Processing cooccurrent variants for sample {}".format(sample_id))

        # the order by position is important to ensure we store only half the matrix and the same half of the matrix
        variants = queries.get_variants_by_sample(sample_id)
        failed_variants = []

        # process all pairwise combinations without repetitions including the diagoonal
        for (variant_one, variant_two) in list(combinations(variants, 2)) + list(zip(variants, variants)):
            try:
                variant_cooccurrence = queries.get_variant_cooccurrence(variant_one, variant_two)
                if variant_cooccurrence is None:
                    variant_cooccurrence = VariantCooccurrence(
                        variant_id_one=variant_one.variant_id,
                        variant_id_two=variant_two.variant_id,
                        count=1
                    )
                    session.add(variant_cooccurrence)
                    session.commit()
                else:
                    # NOTE: it is important to increase the counter like this to avoid race conditions
                    # the increase happens in the database server and not in python
                    # see https://stackoverflow.com/questions/2334824/how-to-increase-a-counter-in-sqlalchemy
                    variant_cooccurrence.count = VariantCooccurrence.count + 1
            except IntegrityError:
                session.rollback()
                failed_variants.append((variant_one, variant_two))

        # tries again the failed variants as these are expected to be there now
        for (variant_one, variant_two) in failed_variants:
            variant_cooccurrence = queries.get_variant_cooccurrence(variant_one, variant_two)
            if variant_cooccurrence is None:
                raise CooccurrenceMatrixException("Some cooccurrent variants failed to be persisted twice")
            variant_cooccurrence.count = VariantCooccurrence.count + 1



import unittest

from covigator.configuration import Configuration
from covigator.pipeline.ena_pipeline import Pipeline
from covigator.pipeline.gisaid_pipeline import GisaidPipeline

class PipelineTest(unittest.TestCase):

    def test_pipeline_run(self):
        fastq1 = "SRR11140748_R1.fastq.gz"
        fastq2 = "SRR11140748_R2.fastq.gz"
        vcf_file = "expected_snpeff.vcf"
        p = Pipeline(config=Configuration())
    
        # TODO: Implement Unit Test here
        # Maybe do diff as assertion?    
        #self.assertEqual(p.run(fastq1=fastq1, fastq2=fastq2), vcf_file)
        #with self.assertRaises(CovigatorPipelineError):
            #vcf_file = p.run(fastq1=fastq1, fastq2=fastq2)
        vcf, qc = p.run(run_accession="test", fastq1=fastq1, fastq2=fastq2)
        self.assertEqual(open(vcf).read(), open(vcf_file).read())


class GisaidPipelineTest(unittest.TestCase):
    
    def test_pipeline_run(self):
        run_accession = "EPI_ISL_417140"
        vcf_file = "gisaid.vcf"

        p = GisaidPipeline(config=Configuration())

        self.assertEqual(open(p.run(run_accession=run_accession)).read(), open(vcf_file).read())

if __name__ == '__main__':
    unittest.main()



import logging
import os
import logzero
from logzero import logger
from covigator.exceptions import CovigatorDashBoardInitialisationError


class Configuration:

    # file system storage
    ENV_COVIGATOR_STORAGE_FOLDER = "COVIGATOR_STORAGE_FOLDER"
    ENV_COVIGATOR_TEMP_FOLDER = "COVIGATOR_TEMP_FOLDER"
    # database
    ENV_COVIGATOR_DB_PORT = "COVIGATOR_DB_PORT"
    ENV_COVIGATOR_DB_PASSWORD = "COVIGATOR_DB_PASSWORD"
    ENV_COVIGATOR_DB_USER = "COVIGATOR_DB_USER"
    ENV_COVIGATOR_DB_NAME = "COVIGATOR_DB_NAME"
    ENV_COVIGATOR_DB_HOST = "COVIGATOR_DB_HOST"
    ENV_COVIGATOR_DB_POOL_SIZE = "COVIGATOR_DB_POOL_SIZE"
    ENV_COVIGATOR_DB_MAX_OVERFLOW = "COVIGATOR_DB_MAX_OVERFLOW"
    ENV_COVIGATOR_TABLE_VERSION = "COVIGATOR_TABLE_VERSION"
    # dashboard
    ENV_COVIGATOR_DASHBOARD_HOST = "COVIGATOR_DASHBOARD_HOST"
    ENV_COVIGATOR_DASHBOARD_PORT = "COVIGATOR_DASHBOARD_PORT"
    # logs
    ENV_COVIGATOR_DASHBOARD_LOG_FILE = "COVIGATOR_DASHBOARD_LOG_FILE"
    ENV_COVIGATOR_PROCESSOR_LOG_FILE = "COVIGATOR_PROCESSOR_LOG_FILE"
    ENV_COVIGATOR_ACCESSOR_LOG_FILE = "COVIGATOR_ACCESSOR_LOG_FILE"
    # pipeline
    ENV_COVIGATOR_NEXTFLOW = "COVIGATOR_NEXTFLOW"
    ENV_COVIGATOR_WORKFLOW = "COVIGATOR_WORKFLOW"
    ENV_COVIGATOR_FORCE_PIPELINE = "COVIGATOR_FORCE_PIPELINE"
    ENV_COVIGATOR_WORKFLOW_CPUS = "COVIGATOR_WORKFLOW_CPUS"
    ENV_COVIGATOR_WORKFLOW_MEMORY = "COVIGATOR_WORKFLOW_MEMORY"
    ENV_COVIGATOR_BATCH_SIZE = "COVIGATOR_BATCH_SIZE"

    # references
    ENV_COVIGATOR_REF_FASTA = "COVIGATOR_REF_FASTA"
    ENV_COVIGATOR_GENE_ANNOTATIONS = "COVIGATOR_GENE_ANNOTATIONS"
    # dask
    ENV_COVIGATOR_DASK_PORT = "COVIGATOR_DASK_PORT"

    def __init__(self):
        # local storage
        self.storage_folder = os.getenv(self.ENV_COVIGATOR_STORAGE_FOLDER, "/data/covigator")

        # database
        self.db_host = os.getenv(self.ENV_COVIGATOR_DB_HOST, "0.0.0.0")
        self.db_name = os.getenv(self.ENV_COVIGATOR_DB_NAME, "covigator")
        self.db_user = os.getenv(self.ENV_COVIGATOR_DB_USER, "covigator")
        self.db_password = os.getenv(self.ENV_COVIGATOR_DB_PASSWORD, "covigator")
        self.db_port = os.getenv(self.ENV_COVIGATOR_DB_PORT, "5432")
        self.db_pool_size = int(os.getenv(self.ENV_COVIGATOR_DB_POOL_SIZE, 5))
        self.db_max_overflow = int(os.getenv(self.ENV_COVIGATOR_DB_MAX_OVERFLOW, 10))
        self.db_table_version = os.environ.get(self.ENV_COVIGATOR_TABLE_VERSION, "")
        self.force_pipeline = os.environ.get(self.ENV_COVIGATOR_FORCE_PIPELINE, False)

        # dashboard
        self.dash_host = os.getenv(self.ENV_COVIGATOR_DASHBOARD_HOST, "0.0.0.0")
        try:
            self.dash_port = int(os.getenv(self.ENV_COVIGATOR_DASHBOARD_PORT, "8050"))
        except ValueError as e:
            raise CovigatorDashBoardInitialisationError("The port needs to be a numeric value. " + str(e))

        # dask
        self.dask_port = os.getenv(self.ENV_COVIGATOR_DASK_PORT, "50218")

        # logs
        self.logfile_dash = os.getenv(self.ENV_COVIGATOR_DASHBOARD_LOG_FILE)
        self.logfile_accesor = os.getenv(self.ENV_COVIGATOR_ACCESSOR_LOG_FILE)
        self.logfile_processor = os.getenv(self.ENV_COVIGATOR_PROCESSOR_LOG_FILE)

        # references
        self.reference_genome = os.getenv(self.ENV_COVIGATOR_REF_FASTA)
        self.reference_gene_annotations = os.getenv(self.ENV_COVIGATOR_GENE_ANNOTATIONS)

        self.nextflow = os.getenv(self.ENV_COVIGATOR_NEXTFLOW, "nextflow")
        self.workflow = os.getenv(self.ENV_COVIGATOR_WORKFLOW, "tron-bioinformatics/covigator-ngs-pipeline -r v0.3.0")
        self.workflow_cpus = os.getenv(self.ENV_COVIGATOR_WORKFLOW_CPUS, "1")
        self.workflow_memory = os.getenv(self.ENV_COVIGATOR_WORKFLOW_MEMORY, "3g")
        try:
            self.batch_size = int(os.getenv(self.ENV_COVIGATOR_BATCH_SIZE, "1000"))
        except ValueError as e:
            raise CovigatorDashBoardInitialisationError("The batch size needs to be a numeric value. " + str(e))
        # NOTE: the defaults are already set in the workflow config
        self.temp_folder = os.getenv(self.ENV_COVIGATOR_TEMP_FOLDER, "/data/covigator-tmp")

        self.log_configuration()

    def log_configuration(self):
        logger.info("Configuration")
        for k, v in self.__dict__.items():
            logger.info("{}={}".format(k, v))


def initialise_logs(logfile, sample_id: str = None):
    if logfile is not None:
        logzero.logfile(logfile)
    logzero.loglevel(logging.INFO)
    if sample_id is not None:
        logzero.formatter(logging.Formatter(
            '[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d ' + sample_id + '] %(message)s'))
    else:
        logzero.formatter(logging.Formatter(
            '[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d] %(message)s'))



import lz4.frame
import base64


def compress_sequence(sequence):
    return base64.b64encode(lz4.frame.compress(sequence.encode('utf-8'))).decode()


def decompress_sequence(s):
    return lz4.frame.decompress(base64.b64decode(s)).decode('utf-8')




from unittest import TestCase

from dash_core_components import Markdown, Graph
from covigator.dashboard.figures.samples import SampleFigures
from covigator.dashboard.figures.variants import VariantsFigures
from covigator.database.database import Database
from faker import Faker

from covigator.database.model import VariantType
from covigator.database.precomputed import Precomputer
from covigator.database.queries import Queries
from covigator.tests.unit_tests.faked_objects import FakeConfiguration
from covigator.tests.unit_tests.mocked import get_mocked_ena_sample, get_mocked_variant, get_mocked_variant_observation, \
    mock_samples_and_variants


class FiguresTests(TestCase):

    def setUp(self) -> None:
        # intialise database
        self.database = Database(test=True, config=FakeConfiguration())
        self.session = self.database.get_database_session()
        self.sample_figures = SampleFigures(queries=Queries(session=self.session))
        self.variants_figures = VariantsFigures(queries=Queries(session=self.session))
        self.faker = Faker()

    def test_samples_by_country(self):
        # populates the ENA samples tables
        for _ in range(100):
            sample_ena, sample, job = get_mocked_ena_sample(faker=self.faker)
            self.session.add_all([sample_ena, sample, job])
        self.session.commit()
        figure = self.sample_figures.get_accumulated_samples_by_country_plot(min_samples=0)
        self.assertIsNotNone(figure)
        self.assertTrue(len(figure) == 2)
        self.assertIsInstance(figure[0], Graph)
        self.assertIsInstance(figure[1], Markdown)

    def test_samples_by_country_no_data(self):
        figure = self.sample_figures.get_accumulated_samples_by_country_plot()
        self.assertIsInstance(figure, Markdown)

    def test_variants_per_sample(self):
        # populates the ENA samples tables
        mock_samples_and_variants(session=self.session, faker=self.faker, num_samples=100)
        Precomputer(session=self.session).load_counts_variants_per_sample()
        figure = self.sample_figures.get_variants_per_sample_plot()
        self.assertIsNotNone(figure)
        self.assertTrue(len(figure) == 2)
        self.assertIsInstance(figure[0], Graph)
        self.assertIsInstance(figure[1], Markdown)

    def test_variants_per_sample_no_data(self):
        figure = self.sample_figures.get_variants_per_sample_plot()
        self.assertIsInstance(figure, Markdown)

    def test_substitutions(self):
        # populates the ENA samples tables
        mock_samples_and_variants(session=self.session, faker=self.faker, num_samples=100)
        Precomputer(session=self.session).load_count_substitutions()
        figure = self.sample_figures.get_substitutions_plot(variant_types=[VariantType.SNV.name])
        self.assertIsNotNone(figure)
        self.assertTrue(len(figure) == 2)
        self.assertIsInstance(figure[0], Graph)
        self.assertIsInstance(figure[1], Markdown)

    def test_substitutions_no_data(self):
        figure = self.sample_figures.get_substitutions_plot(variant_types=[VariantType.SNV.name])
        self.assertIsInstance(figure, Markdown)

    def test_needle_plot_no_data(self):
        figure = self.variants_figures.get_variants_plot(gene_name="S", selected_variants=None, bin_size=50)
        self.assertIsInstance(figure, Markdown)



from unittest import TestCase
from dask.distributed import Client

from covigator.configuration import Configuration
from covigator.database.database import Database
from covigator.database.model import Log, DataSource, CovigatorModule
from covigator.processor.ena_processor import EnaProcessor
from faker import Faker
from covigator.processor.gisaid_processor import GisaidProcessor
from covigator.tests.unit_tests.faked_objects import FakeConfiguration


class ProcessorTests(TestCase):

    def setUp(self) -> None:
        # intialise database
        config = FakeConfiguration()
        self.database = Database(test=True, config=config)
        self.session = self.database.get_database_session()
        self.ena_processor = EnaProcessor(
            database=self.database, dask_client=Client(n_workers=int(1), threads_per_worker=1), config=config)
        self.gisaid_processor = GisaidProcessor(
            database=self.database, dask_client=Client(n_workers=int(1), threads_per_worker=1), config=config)
        self.faker = Faker()

    def test_no_ena_jobs(self):
        self.ena_processor.process()
        self.assertEqual(self.session.query(Log).count(), 1)
        log = self.session.query(Log).first()
        self.assertIsNotNone(log.start)
        self.assertIsNotNone(log.end)
        self.assertEqual(log.source, DataSource.ENA)
        self.assertEqual(log.module, CovigatorModule.PROCESSOR)
        self.assertEqual(log.processed, 0)
        data = log.data
        self.assertEqual(data.get("processed"), 0)

    def test_no_gisaid_jobs(self):
        self.gisaid_processor.process()
        self.assertEqual(self.session.query(Log).count(), 1)
        log = self.session.query(Log).first()
        self.assertIsNotNone(log.start)
        self.assertIsNotNone(log.end)
        self.assertEqual(log.source, DataSource.GISAID)
        self.assertEqual(log.module, CovigatorModule.PROCESSOR)
        self.assertEqual(log.processed, 0)
        data = log.data
        self.assertEqual(data.get("processed"), 0)



from datetime import date
import os
from unittest import TestCase
import pkg_resources
from Bio import SeqIO

import covigator
from covigator.accessor.gisaid_accessor import GisaidAccessor
from covigator.configuration import Configuration
from covigator.database.database import Database
from covigator.database.model import SampleGisaid, Sample, JobGisaid
from covigator.misc.compression import decompress_sequence
from covigator.pipeline.gisaid_pipeline import GisaidPipeline
from covigator.tests.unit_tests.faked_objects import FakeConfiguration


class GisaidAccessorTests(TestCase):

    def setUp(self) -> None:
        self.config = FakeConfiguration()
        dirname=os.path.dirname
        self.input_fasta = os.path.join(dirname(dirname(__file__)), "resources", "gisaid_allprot0526.fasta")
        self.input_metadata = os.path.join(dirname(dirname(__file__)), "resources", "metadata_2021_05_26.csv")
        self.database = Database(test=True, config=self.config)

    def test_gisaid_accessor(self):
        GisaidAccessor(input_fasta=self.input_fasta, input_metadata=self.input_metadata, database=self.database).access()
        session = self.database.get_database_session()
        self.assertEqual(session.query(SampleGisaid).count(), 1)
        self.assertEqual(session.query(Sample).count(), 1)
        self.assertEqual(session.query(JobGisaid).count(), 1)
        samples = session.query(SampleGisaid).all()
        for s in samples:
            self.assertEqual(s.country, "China")
            self.assertEqual(s.country_alpha_2, "CN")
            self.assertEqual(s.country_alpha_3, "CHN")
            self.assertEqual(s.continent, "Asia")
            self.assertEqual(s.continent_alpha_2, "AS")
            self.assertEqual(s.host, "human")
            # checks that every sequence after decompression can be found in the fasta file
            for g, seq in s.sequence.items():
                decompressed_sequence = decompress_sequence(seq)
                found = False
                for record in SeqIO.parse(self.input_fasta, "fasta"):
                    if str(record.seq) == decompressed_sequence:
                        found = True
                self.assertTrue(found)



from unittest import TestCase
import pkg_resources
import covigator.tests
from covigator.database.database import Database
from covigator.database.model import Variant, VariantObservation, Sample, DataSource, SubclonalVariantObservation, \
    SampleGisaid, SampleEna
from covigator.exceptions import CovigatorExcludedSampleTooManyMutations
from covigator.pipeline.vcf_loader import VcfLoader
from covigator.tests.unit_tests.faked_objects import FakeConfiguration


class VcfLoaderTests(TestCase):

    def setUp(self) -> None:
        self.session = Database(test=True, config=FakeConfiguration()).get_database_session()
        self.sample = Sample(id="TEST1", source=DataSource.ENA)
        sample_ena = SampleEna(run_accession="TEST1", fastq_ftp="something", fastq_md5="else", num_fastqs=2)
        self.sample2 = Sample(id="TEST2", source=DataSource.GISAID)
        sample_gisaid = SampleGisaid(run_accession="TEST2")
        self.session.add_all([self.sample, self.sample2, sample_ena, sample_gisaid])
        self.session.commit()

    def test_vcf_loader(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff.vcf")
        VcfLoader().load(vcf_file, self.sample, self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 3)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)
        variant = self.session.query(Variant).first()
        self.assertEqual(variant.chromosome, "MN908947.3")
        self.assertEqual(variant.position, 23403)
        self.assertEqual(variant.reference, "A")
        self.assertEqual(variant.alternate, "G")
        self.assertEqual(variant.gene_name, "S")
        variant_observation = self.session.query(VariantObservation).first()
        self.assertEqual(variant_observation.sample, self.sample.id)
        self.assertEqual(variant_observation.source, self.sample.source)
        self.assertEqual(variant_observation.chromosome, "MN908947.3")
        self.assertEqual(variant_observation.position, 23403)
        self.assertEqual(variant_observation.reference, "A")
        self.assertEqual(variant_observation.alternate, "G")
        self.assertEqual(variant_observation.dp, 38)
        self.assertEqual(variant_observation.vaf, 1.0)
        self.assertEqual(self.session.query(SubclonalVariantObservation).count(), 2)

    def test_vcf_loader_without_dp4(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff_without_dp4.vcf")
        VcfLoader().load(vcf_file, self.sample, self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 1)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)

    def test_vcf_loader_without_annotation(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff_without_annotation.vcf")
        VcfLoader().load(vcf_file, self.sample, self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 1)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)

    def test_vcf_loader_with_insertion(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff_with_insertion.vcf")
        VcfLoader().load(vcf_file, self.sample, self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 1)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)

    def test_vcf_loader_with_deletion(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff_with_deletion.vcf")
        VcfLoader().load(vcf_file, self.sample, self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 1)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)

    def test_vcf_loader_gisaid(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff.vcf")
        VcfLoader().load(vcf_file, self.sample2, self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 3)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)
        variant = self.session.query(Variant).first()
        self.assertEqual(variant.chromosome, "MN908947.3")
        self.assertEqual(variant.position, 23403)
        self.assertEqual(variant.reference, "A")
        self.assertEqual(variant.alternate, "G")
        self.assertEqual(variant.gene_name, "S")
        variant_observation = self.session.query(VariantObservation).first()
        self.assertEqual(variant_observation.sample, self.sample2.id)
        self.assertEqual(variant_observation.source, self.sample2.source)
        self.assertEqual(variant_observation.chromosome, "MN908947.3")
        self.assertEqual(variant_observation.position, 23403)
        self.assertEqual(variant_observation.reference, "A")
        self.assertEqual(variant_observation.alternate, "G")
        self.assertEqual(self.session.query(SubclonalVariantObservation).count(), 2)

    def test_too_many_mutations(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff.vcf")
        self.session.commit()
        try:
            VcfLoader().load(vcf_file, self.sample2, self.session, max_snvs=0)
            self.assertTrue(False)
        except CovigatorExcludedSampleTooManyMutations:
            self.assertTrue(True)



import functools

import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import plotly

from covigator.dashboard.figures.figures import PLOTLY_CONFIG, MARGIN, TEMPLATE
from covigator.dashboard.tabs import get_mini_container, print_number, print_date
from covigator.database.model import DataSource, SAMPLE_GISAID_TABLE_NAME
from covigator.database.queries import Queries
import pandas as pd
import plotly.express as px


@functools.lru_cache()
def get_tab_dataset_gisaid(queries: Queries):

    count_samples = queries.count_samples(source=DataSource.GISAID.name)
    count_variants = queries.count_variant_observations(source=DataSource.GISAID.name)
    date_of_first_gisaid_sample = queries.get_date_of_first_sample(source=DataSource.GISAID)
    date_of_most_recent_gisaid_sample = queries.get_date_of_most_recent_sample(source=DataSource.GISAID)

    return dbc.Card(
        dbc.CardBody(
            children=[
                dcc.Markdown("""
                    The GISAID dataset was manually downloaded from the site https://www.gisaid.org/.
                    DNA assemblies and metadata were matched together and variant calling was done after performing 
                    a global alignment to the reference genome.
                    """, style={"font-size": 16}),
                html.Br(),
                html.Div(
                    html.Span(
                        children=[
                            get_mini_container(
                                title="Samples",
                                value=print_number(count_samples)),
                            get_mini_container(
                                title="Variant calls",
                                value=print_number(count_variants)),
                            get_mini_container(
                                title="First sample",
                                value=print_date(date_of_first_gisaid_sample)),
                            get_mini_container(
                                title="Latest sample",
                                value=print_date(date_of_most_recent_gisaid_sample))
                            ])
                ),
                get_dataset_gisaid_tab_graphs(queries=queries, count_samples=count_samples)
        ])
    )


def get_dataset_gisaid_tab_graphs(queries: Queries, count_samples):

    return html.Div(
        children=[
            html.Br(),
            html.Div(children=[
                html.Div(children=get_plot_coverage(queries),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
                html.Div(children=get_plot_bad_bases_ratio(queries, count_samples),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
            ]),
        ])


def get_plot_coverage(queries: Queries):
    sql_query = """
    select count(*), (sequence_length::float / 29903 * 100)::int as coverage
    from {table} 
    where finished
    group by coverage
    """.format(table=SAMPLE_GISAID_TABLE_NAME)
    data = pd.read_sql_query(sql_query, queries.session.bind)
    fig = px.bar(data_frame=data, x="coverage", y="count", log_y=True, color="count",
                 color_continuous_scale=plotly.colors.sequential.Brwnyl)
    fig.update_layout(
        margin=MARGIN,
        template=TEMPLATE,
        yaxis={'title': "Num. of samples"},
        xaxis={'title': "Horizontal coverage (%)"},
        legend={'title': None}
    )
    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Horizontal coverage (%)**
        
        Horizontal coverage is estimated as the sequence length divided by the reference genome length.
        Some samples have thus an horizontal coverage larger than 100 %.
        Samples covering less than 20% of the whole genome are excluded. 
        """)
    ]


def get_plot_bad_bases_ratio(queries: Queries, count_samples):
    sql_query = """
    select count(*), ((count_n_bases + count_ambiguous_bases)::float / sequence_length * 100)::int as bad_bases_ratio
    from {table} 
    where finished
    group by bad_bases_ratio
    """.format(table=SAMPLE_GISAID_TABLE_NAME)
    data = pd.read_sql_query(sql_query, queries.session.bind)
    fig = px.bar(data_frame=data, x="bad_bases_ratio", y="count", log_y=True, color="count",
                 color_continuous_scale=plotly.colors.sequential.Brwnyl)
    fig.update_layout(
        margin=MARGIN,
        template=TEMPLATE,
        yaxis={'title': "Num. of samples"},
        xaxis={'title': "Ratio of N and ambiguous bases (%)"},
        legend={'title': None}
    )


    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Ratio of N and ambiguous bases (%)**

        The ratio of N and ambiguous bases over the whole sequence length measures the quality of the assembly sequence.
        {} % of samples have a ratio <= 5 %.
        Samples with a ratio higher than 0.2 are excluded.
        All variant calls that contain a N or an ambiguous base are filtered out.
        """.format(round(float(data[data["bad_bases_ratio"] <= 5]["count"].sum()) / count_samples), 1))
    ]


from flask import Flask
from covigator.configuration import Configuration
from covigator.dashboard.dashboard import Dashboard
from logzero import logger

dashboard = Dashboard(config=Configuration())
app = dashboard.get_application()
server = Flask(__name__)
app.init_app(server)

if __name__ == '__main__':
    logger.info("Starting...")
    server.run()
    logger.info("Running...")



from datetime import datetime
import dash_bootstrap_components as dbc

MONTH_PATTERN = "%Y-%m"
MISSING_VALUE = "-"


def print_date(date: datetime.date):
    return str(date) if date is not None else MISSING_VALUE


def print_number(value):
    return '{:,}'.format(value)


def get_mini_container(title, value, font_size=20):
    return dbc.Button(
        [title + " ", dbc.Badge(value, color="light", className="ml-1")],
        color="warning", size='lg',  style={"margin-left": "20px", "font-size": font_size})



from setuptools import find_packages, setup
import covigator


# parses requirements from file
with open("requirements.txt") as f:
    required = f.read().splitlines()

with open("README.md", "r") as f:
    long_description = f.read()

# Build the Python package
setup(
    name="covigator",
    version=covigator.VERSION,
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "covigator-ena-accessor=covigator.command_line:ena_accessor",
            "covigator-gisaid-accessor=covigator.command_line:gisaid_accessor",
            "covigator-processor=covigator.command_line:processor",
            "covigator-dashboard=covigator.dashboard.dashboard:main",
            "covigator-pipeline=covigator.command_line:pipeline",
            "covigator-precompute=covigator.command_line:precompute_queries"
        ],
    },
    author_email="patrick.sorn@tron-mainz.de",
    author="TRON - Translational Oncology at the University Medical Center of the Johannes Gutenberg University Mainz "
           "- Computational Medicine group",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    requires=[],
    # NOTE: always specify versions to ensure build reproducibility
    # NOTE2: sklearn==0.19.0 is a hidden dependency as it is required by Classifier.pickle
    install_requires=required,
    setup_requires=[],
    classifiers=[
        "Development Status :: 3 - Alpha",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Programming Language :: Python :: 3",
    ],
)



import json
from datetime import datetime

import pandas as pd

from covigator.configuration import Configuration
from covigator.database.queries import Queries
from covigator.exceptions import CovigatorErrorProcessingCoverageResults, CovigatorExcludedSampleBadQualityReads, \
    CovigatorExcludedSampleNarrowCoverage
from covigator.misc import backoff_retrier
from covigator.database.model import JobStatus, JobEna, Sample, DataSource
from covigator.database.database import Database
from logzero import logger
from dask.distributed import Client
from covigator.processor.abstract_processor import AbstractProcessor
from covigator.pipeline.cooccurrence_matrix import CooccurrenceMatrix
from covigator.pipeline.downloader import Downloader
from covigator.pipeline.ena_pipeline import Pipeline
from covigator.pipeline.vcf_loader import VcfLoader

NUMBER_RETRIES_DOWNLOADER = 10


class EnaProcessor(AbstractProcessor):

    def __init__(self, database: Database, dask_client: Client, config: Configuration):
        logger.info("Initialising ENA processor")
        super().__init__(database, dask_client, DataSource.ENA, config)

    def _process_run(self, run_accession: str):
        """
        Launches all jobs and returns the futures for the final job only
        """
        # NOTE: here we set the priority of each step to ensure a depth first processing
        future = self.dask_client.submit(EnaProcessor.job, self.config, run_accession, priority=1)
        return future

    @staticmethod
    def job(config: Configuration, run_accession):
        return EnaProcessor.run_job(
            config, run_accession, start_status=JobStatus.QUEUED, end_status=JobStatus.FINISHED,
            error_status=JobStatus.FAILED_PROCESSING, data_source=DataSource.ENA,
            function=EnaProcessor.run_all)

    @staticmethod
    def run_all(job: JobEna, queries: Queries, config: Configuration):
        EnaProcessor.download(job=job, queries=queries, config=config)
        EnaProcessor.run_pipeline(job=job, queries=queries, config=config)
        EnaProcessor.load(job=job, queries=queries, config=config)
        EnaProcessor.compute_cooccurrence(job=job, queries=queries, config=config)

    @staticmethod
    def download(job: JobEna, queries: Queries, config: Configuration):
        # ensures that the download is done with retries, even after MD5 check sum failure
        downloader = Downloader(config=config)
        download_with_retries = backoff_retrier.wrapper(downloader.download, NUMBER_RETRIES_DOWNLOADER)
        sample_ena = queries.find_sample_by_accession(job.run_accession, source=DataSource.ENA)
        paths = download_with_retries(sample_ena=sample_ena)
        job.fastq_path = paths
        job.downloaded_at = datetime.now()

    @staticmethod
    def run_pipeline(job: JobEna, queries: Queries, config: Configuration):
        fastq1, fastq2 = job.get_fastq1_and_fastq2()
        vcf_path, qc_path, vertical_coverage_path, horizontal_coverage_path = Pipeline(config=config)\
            .run(run_accession=job.run_accession, fastq1=fastq1, fastq2=fastq2)
        job.analysed_at = datetime.now()
        job.vcf_path = vcf_path
        job.qc_path = qc_path
        job.qc = json.load(open(qc_path))
        job.horizontal_coverage_path = horizontal_coverage_path
        job.vertical_coverage_path = vertical_coverage_path
        EnaProcessor.load_coverage_results(horizontal_coverage_path, job)

    @staticmethod
    def load_coverage_results(horizontal_coverage_path, job):
        try:
            data = pd.read_csv(horizontal_coverage_path, sep="\t")
            job.mean_depth = float(data.meandepth.loc[0])
            job.mean_base_quality = float(data.meanbaseq.loc[0])
            job.mean_mapping_quality = float(data.meanmapq.loc[0])
            job.num_reads = int(data.numreads.loc[0])
            job.num_reads = int(data.numreads.loc[0])
            job.covered_bases = int(data.covbases.loc[0])
            job.coverage = float(data.coverage.loc[0])
        except Exception as e:
            raise CovigatorErrorProcessingCoverageResults(e)

    @staticmethod
    def load(job: JobEna, queries: Queries, config: Configuration):
        if job.mean_mapping_quality < 10 or job.mean_base_quality < 10:
            raise CovigatorExcludedSampleBadQualityReads("Mean MQ: {}; mean BCQ: {}".format(
                job.mean_mapping_quality, job.mean_base_quality))
        if job.coverage < 20.0:
            raise CovigatorExcludedSampleNarrowCoverage("Horizontal coverage {} %".format(job.coverage))
        VcfLoader().load(
            vcf_file=job.vcf_path, sample=Sample(id=job.run_accession, source=DataSource.ENA), session=queries.session)
        job.loaded_at = datetime.now()

    @staticmethod
    def compute_cooccurrence(job: JobEna, queries: Queries, config: Configuration):
        CooccurrenceMatrix().compute(
            sample=Sample(id=job.run_accession, source=DataSource.ENA), session=queries.session)
        job.cooccurrence_at = datetime.now()



import os
from pathlib import Path
from logzero import logger
from covigator.configuration import Configuration
from covigator.pipeline.runner import run_command


class Pipeline:

    def __init__(self, config: Configuration):
        self.config = config
        assert self.config.reference_genome is not None and os.path.exists(self.config.reference_genome), \
            "Please configure the reference genome in the variable {}".format(self.config.ENV_COVIGATOR_REF_FASTA)

    def run(self, run_accession: str, fastq1: str, fastq2: str = None):

        logger.info("Processing {} and {}".format(fastq1, fastq2))
        sample_data_folder = Path(fastq1).parent

        logger.info("Sample data folder: {}".format(sample_data_folder))
        output_vcf = os.path.join(
            self.config.storage_folder, run_accession,
            "{name}.lofreq.normalized.annotated.vcf.gz".format(name=run_accession))
        output_qc = os.path.join(
            self.config.storage_folder, run_accession, "{name}.fastp_stats.json".format(name=run_accession))
        output_vertical_coverage = os.path.join(
            self.config.storage_folder, run_accession, "{name}.depth.tsv".format(name=run_accession))
        output_horizontal_coverage = os.path.join(
            self.config.storage_folder, run_accession, "{name}.coverage.tsv".format(name=run_accession))

        if not os.path.exists(output_vcf) \
                or not os.path.exists(output_qc) \
                or not os.path.exists(output_horizontal_coverage) \
                or not os.path.exists(output_vertical_coverage) \
                or self.config.force_pipeline:

            command = "{nextflow} run {workflow} " \
                      "--fastq1 {fastq1} {fastq2} --output {output_folder} --name {name} " \
                      "--cpus {cpus} --memory {memory} " \
                      "-profile conda -offline -work-dir {work_folder} -with-trace {trace_file}".format(
                nextflow=self.config.nextflow,
                fastq1=fastq1,
                fastq2="--fastq2 " + fastq2 if fastq2 else "",
                output_folder=self.config.storage_folder,
                name=run_accession,
                work_folder=self.config.temp_folder,
                workflow=self.config.workflow,
                trace_file=os.path.join(sample_data_folder, "nextflow_traces.txt"),
                cpus=self.config.workflow_cpus,
                memory=self.config.workflow_memory
            )
            run_command(command, sample_data_folder)

        return output_vcf, output_qc, output_vertical_coverage, output_horizontal_coverage



import dash_html_components as html

import covigator


def get_footer():
    tron_logo = "/assets/tron_logo_no_bg.png"
    return html.Footer(
        [
            html.Div(
                children=[
                    html.Br(),
                    html.P("CoVigator {} © 2021 TRON. All Rights Reserved".format(covigator.VERSION)),
                    html.A(html.Img(src=tron_logo, id="tron-logo"), href="https://tron-mainz.de"),
                    html.P("The CoVigator project is enabled by Intel Corporation servers."),
                    html.P([
                        html.A("DATA PROTECTION", href="https://tron-mainz.de/data-protection/"), " | ",
                        html.A("IMPRINT", href="https://tron-mainz.de/imprint/")])
                ],
                # this bit makes sure the footer sticks at the bottom
                #style={"position": "relative", "bottom": "0", "width": "100%", "overflow": "hidden", "height": "120px"}
            )
        ])



from logzero import logger
import time
import subprocess
from covigator.exceptions import CovigatorPipelineError


def run_command(command, temporary_folder):
    start = time.time()
    p = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=temporary_folder, shell=True)
    stdoutdata, stderrdata = p.communicate()
    logger.info("Finished in {} secs command: '{}'".format(time.time() - start, command))
    if p.returncode != 0:
        error_message = decode(stderrdata)
        logger.error(error_message)
        raise CovigatorPipelineError("Error executing pipeline command: {}\n{}".format(command, error_message))


def decode(data):
    return data.decode("utf8")



#!/usr/bin/env python
import os
from Bio import SeqIO
from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from logzero import logger
from covigator.configuration import Configuration
from covigator.database.model import SampleGisaid
from covigator.exceptions import CovigatorExcludedAssemblySequence
from covigator.misc.compression import decompress_sequence
from covigator.pipeline.runner import run_command

MINIMUM_SEQUENCE_SIZE = 1000


class GisaidPipeline:

    def __init__(self, config: Configuration):
        self.config = config

    def run(self, sample: SampleGisaid):
        logger.info("Processing {}".format(sample.run_accession))
        sample_name = sample.run_accession.replace("/", "_").replace(" ", "-").replace("'", "-").replace("$", "")
        sample_data_folder = os.path.join(self.config.storage_folder, sample_name)
        output_vcf = os.path.join(
            sample_data_folder,
            "{name}.assembly.normalized.annotated.vcf.gz".format(name=sample_name))
        input_fasta = os.path.join(sample_data_folder, "{}.fasta".format(sample_name))

        if not os.path.exists(output_vcf) or self.config.force_pipeline:

            # writes the FASTA sequence into a file
            _, sequence = list(sample.sequence.items())[0]
            os.makedirs(sample_data_folder, exist_ok=True)
            decompressed_sequence = decompress_sequence(sequence)

            # excludes too small sequences
            if len(decompressed_sequence) < MINIMUM_SEQUENCE_SIZE:
                raise CovigatorExcludedAssemblySequence

            with open(input_fasta, "w+") as output:
                record = SeqRecord(
                    seq=Seq(decompressed_sequence, IUPAC.ambiguous_dna),
                    id=sample_name)
                SeqIO.write(record, output, "fasta")

            command = "{nextflow} run {workflow} " \
                      "--fasta {fasta} --output {output_folder} --name {name} " \
                      "--cpus {cpus} --memory {memory} " \
                      "-profile conda -offline -work-dir {work_folder} -with-trace {trace_file}".format(
                nextflow=self.config.nextflow,
                fasta=input_fasta,
                output_folder=self.config.storage_folder,
                name=sample_name,
                work_folder=self.config.temp_folder,
                workflow=self.config.workflow,
                trace_file=os.path.join(sample_data_folder, "nextflow_traces.txt"),
                cpus=self.config.workflow_cpus,
                memory=self.config.workflow_memory
            )
            run_command(command, sample_data_folder)

        return output_vcf



import os
import pathlib
import urllib.request as request
from contextlib import closing
import shutil
import hashlib
from logzero import logger
from covigator.configuration import Configuration
from covigator.database.model import SampleEna, SEPARATOR
import re

FTP_PROTOCOL = "ftp://"
PROTOCOL_REGEX = re.compile('[a-zA-Z0-9]+://')


class CovigatorMD5CheckSumError(Exception):
    pass


class Downloader:

    def __init__(self, config: Configuration):
        self.storage_folder = config.storage_folder
        pathlib.Path(self.storage_folder).mkdir(parents=True, exist_ok=True)
        assert os.path.exists(self.storage_folder), "Storage folder does not exist"

    def download(self, sample_ena: SampleEna) -> str:
        assert sample_ena.run_accession is not None, "A run accession is required as it is part of the file path"
        assert sample_ena.fastq_ftp is not None, "Cannot download empty URLs"
        assert sample_ena.fastq_md5 is not None, "Cannot do MD5 checks without MD5"
        fastqs = sample_ena.get_fastqs_ftp()
        md5s = sample_ena.get_fastqs_md5()
        paths = []
        for url, md5 in zip(fastqs, md5s):
            local_path = self._download_url(sample_ena.run_accession, url, md5)
            paths.append(local_path)
        logger.info("Downloaded {}: {}".format(sample_ena.run_accession, sample_ena.fastq_ftp))
        return SEPARATOR.join(paths)

    def _download_url(self, run_accession, url, md5):
        """
        This method streams a file (probably large) from a URL and stores it in the local file system
        """
        if not re.match(PROTOCOL_REGEX, url):
            url = FTP_PROTOCOL + url
        local_filename = url.split('/')[-1]
        local_folder = os.path.join(self.storage_folder, run_accession)
        local_full_path = os.path.join(local_folder, local_filename)
        # avoids downloading the same files over and over
        if not os.path.exists(local_full_path):
            pathlib.Path(local_folder).mkdir(parents=True, exist_ok=True)
            request.urlretrieve(url, local_full_path)
            self._check_md5(local_full_path, md5)

        return local_full_path

    def _check_md5(self, filepath, md5):
        """
        This method reads a file (probably large) in chunks and computes the MD5 hash iteratively
        """
        file_hash = hashlib.md5()
        with open(filepath, "rb") as f:
            while True:
                chunk = f.read(8192)
                if not chunk:
                    break
                file_hash.update(chunk)
        if file_hash.hexdigest() != md5:
            raise CovigatorMD5CheckSumError("Failed MD5 check for file {} and MD5 {}".format(filepath, md5))
